from __future__ import annotations

import math
from typing import TYPE_CHECKING

import numpy as np

import gamspy as gp
import gamspy.formulations.utils as utils
from gamspy.exceptions import ValidationError
from gamspy.math import dim

if TYPE_CHECKING:
    from gamspy import Parameter, Variable


class Linear:
    """
    Formulation generator for Linear layer in GAMS.

    Parameters
    ----------
    container : Container
        Container that will contain the new variable and equations.
    in_features : int
        Input feature size
    out_features : int
        Output feature size
    bias : bool = True
        Should bias be added after linear transformation, by Default: True
    name_prefix : str | None
        Prefix for generated GAMSPy symbols, by default None which means
        random prefix. Using the same name_prefix in different formulations causes name
        conflicts. Do not use the same name_prefix again.

    Examples
    --------
    >>> import gamspy as gp
    >>> import numpy as np
    >>> from gamspy.math import dim
    >>> m = gp.Container()
    >>> l1 = gp.formulations.Linear(m, 128, 64)
    >>> w = np.random.rand(64, 128)
    >>> b = np.random.rand(64)
    >>> l1.load_weights(w, b)
    >>> x = gp.Variable(m, "x", domain=dim([10, 128]))
    >>> y, set_y = l1(x)
    >>> [d.name for d in y.domain]
    ['DenseDim10_1', 'DenseDim64_1']

    """

    def __init__(
        self,
        container: gp.Container,
        in_features: int,
        out_features: int,
        bias: bool = True,
        name_prefix: str | None = None,
    ):
        if not isinstance(in_features, int) or in_features <= 0:
            raise ValidationError("in_features must be a positive integer")

        if not isinstance(out_features, int) or out_features <= 0:
            raise ValidationError("out_features must be a positive integer")

        if not isinstance(bias, bool):
            raise ValidationError("bias must be a boolean")

        self.container = container
        self.in_features = in_features
        self.out_features = out_features
        self.use_bias = bias
        self._state = 0
        self.weight: Parameter | Variable | None = None
        self.weight_array = None
        self.bias: Parameter | Variable | None = None
        self.bias_array = None

        if name_prefix is None:
            name_prefix = gp.utils._get_unique_name()

        self._name_prefix = name_prefix

    def load_weights(self, weight: np.ndarray, bias: np.ndarray | None = None) -> None:
        """
        Mark Linear as parameter and load weights from NumPy arrays.
        After this is called `make_variable` cannot be called. Use this
        when you already have the weights of your Linear layer.

        Parameters
        ----------
        weight : np.ndarray
               Linear layer weights in shape (out_features x in_features)
        bias : np.ndarray | None
               Linear layer bias in shape (out_features, ), only required when
               bias=True during initialization

        """
        if self._state == 2:
            raise ValidationError(
                "load_weights cannot be used after calling make_variable"
            )

        if self.use_bias is False and bias is not None:
            raise ValidationError(
                "bias must be None since bias was set to False during initialization"
            )

        if self.use_bias is True and bias is None:
            raise ValidationError("bias must be provided")

        if len(weight.shape) != 2:
            raise ValidationError(
                f"expected 2D input for weight (got {len(weight.shape)}D input)"
            )

        expected_shape = (
            self.out_features,
            self.in_features,
        )
        if weight.shape != expected_shape:
            raise ValidationError(f"weight expected to be in shape {expected_shape}")

        if bias is not None:
            if len(bias.shape) != 1:
                raise ValidationError(
                    f"expected 1D input for bias (got {len(bias.shape)}D input)"
                )

            if bias.shape[0] != self.out_features:
                raise ValidationError(
                    f"bias expected to be in shape ({self.out_features},)"
                )

        if self.weight is None:
            self.weight = gp.Parameter(
                self.container,
                name=utils._generate_name("p", self._name_prefix, "weight"),
                domain=dim(expected_shape),
                records=weight,
            )
        else:
            self.weight.setRecords(weight)
        self.weight_array = weight

        if self.use_bias:
            if self.bias is None:
                self.bias = gp.Parameter(
                    self.container,
                    name=utils._generate_name("p", self._name_prefix, "bias"),
                    domain=dim([self.out_features]),
                    records=bias,
                )
            else:
                self.bias.setRecords(bias)

            self.bias_array = bias

        self._state = 1

    def make_variable(self, *, init_weights=False) -> None:
        """
        Mark Linear layer as variable. After this is called `load_weights`
        cannot be called. Use this when you need to learn the weights
        of your linear layer in your optimization model.


        Parameters
        ----------
        init_weights : Optional[bool]
               False by default.
               Whether to initialize weights. It is suggested you set
               this to True unless you want to initialize weights yourself.
               When `init_weights` is set to True, values are initialized from
               :math:`\\mathcal{U}(-\\sqrt{k},\\sqrt{k})`, where :math:`k = 1/in\\_features`.
        """
        if self._state == 1:
            raise ValidationError(
                "make_variable cannot be used after calling load_weights"
            )

        expected_shape = (
            self.out_features,
            self.in_features,
        )

        sk = math.sqrt(1 / self.in_features)
        if self.weight is None:
            self.weight = gp.Variable(
                self.container,
                name=utils._generate_name("v", self._name_prefix, "weight"),
                domain=dim(expected_shape),
            )
            if init_weights:
                self.weight.l[...] = gp.math.uniform(-sk, sk)

        if self.use_bias and self.bias is None:
            self.bias = gp.Variable(
                self.container,
                name=utils._generate_name("v", self._name_prefix, "bias"),
                domain=dim([self.out_features]),
            )
            if init_weights:
                self.bias.l[...] = gp.math.uniform(-sk, sk)

        self._state = 2

    def __call__(
        self, input: gp.Parameter | gp.Variable, propagate_bounds: bool = True
    ) -> tuple[gp.Variable, list[gp.Equation]]:
        """
        Forward pass your input, generate output and equations required for
        calculating the linear transformation. If `propagate_bounds` is True,
        the `input` is of type variable, and `load_weights` was called, then
        the bounds of the input are propagated to the output.

        Parameters
        ----------
        input : gp.Parameter | gp.Variable
                input to the linear layer, must be in shape
                (* x in_features)
        propagate_bounds : bool = True
                If True, propagate bounds of the input to the output.
                Otherwise, the output variable is unbounded.
        """
        if not isinstance(propagate_bounds, bool):
            raise ValidationError("propagate_bounds should be a boolean.")

        if self.weight is None:
            raise ValidationError(
                "You must call load_weights or make_variable first before using the Linear"
            )

        if len(input.domain) == 0:
            raise ValidationError("expected an input with at least 1 dimension")

        if len(input.domain[-1]) != self.in_features:
            raise ValidationError("in_features does not match")

        expr = input @ self.weight.t()

        if self.bias is not None:
            expr = expr + self.bias[expr.domain[-1]]

        out = gp.Variable(
            self.container,
            name=utils._generate_name("v", self._name_prefix, "output"),
            domain=expr.domain,
        )

        set_out = gp.Equation(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "set_output"),
            domain=out.domain,
        )

        set_out[...] = out == expr

        # If propagate_bounds is True, weight is a parameter and input is a variable,
        # we will propagate the bounds of the input to the output
        if propagate_bounds and self._state == 1 and isinstance(input, gp.Variable):
            x_bounds = gp.Parameter(
                self.container,
                name=utils._generate_name("p", self._name_prefix, "input_bounds"),
                domain=dim([2, *input.shape]),
            )
            x_bounds[("0",) + tuple(input.domain)] = input.lo[...]
            x_bounds[("1",) + tuple(input.domain)] = input.up[...]

            # If the bounds are all zeros (None in GAMSPy parameters);
            # we skip matrix multiplication as it will result in zero values
            if x_bounds.records is None:
                out_bounds_array = np.zeros(out.shape)

                if self.use_bias:
                    out_bounds_array = out_bounds_array + self.bias_array

                out_bounds = gp.Parameter(
                    self.container,
                    name=utils._generate_name("p", self._name_prefix, "output_bounds"),
                    domain=dim(out.shape),
                    records=out_bounds_array,
                )
                out.lo[...] = out_bounds
                out.up[...] = out_bounds

                return out, [set_out]

            x_lb, x_ub = x_bounds.toDense()

            # To deal with infinity values in the input bounds, we convert them into complex numbers
            # where if the value is -inf, we convert it to 0 - 1j
            # and if the value is inf, we convert it to 0 + 1j
            x_lb = np.where(x_lb == -np.inf, 0 - 1j, x_lb)
            x_ub = np.where(x_ub == np.inf, 0 + 1j, x_ub)

            # get the positive and negative weights separately
            w_pos = np.maximum(self.weight_array, 0)
            w_neg = np.minimum(self.weight_array, 0)

            lo_out = (x_lb @ w_pos.T) + (x_ub @ w_neg.T)
            up_out = (x_ub @ w_pos.T) + (x_lb @ w_neg.T)

            def _decode_complex_number(z: np.complex128) -> float:
                """
                Decode complex number to real number.
                5 + 0j -> 5
                3 + 1j -> inf
                7 - 3j -> -inf
                """
                # If imaginary part is zero, return real part
                if z.imag == 0:
                    return z.real
                # If imaginary part is positive, return positive infinity
                elif z.imag > 0:
                    return np.inf
                # If imaginary part is negative, return negative infinity
                else:
                    return -np.inf

            lo_out = np.vectorize(_decode_complex_number)(lo_out)
            up_out = np.vectorize(_decode_complex_number)(up_out)

            if self.use_bias:
                lo_out = lo_out + self.bias_array
                up_out = up_out + self.bias_array

            out_bounds_array = np.stack([lo_out, up_out], axis=0)

            out_bounds = gp.Parameter(
                self.container,
                name=utils._generate_name("p", self._name_prefix, "output_bounds"),
                domain=dim([2, *out.shape]),
                records=out_bounds_array,
            )

            out.lo[...] = out_bounds[("0",) + tuple(out.domain)]
            out.up[...] = out_bounds[("1",) + tuple(out.domain)]

        return out, [set_out]



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal, no_type_check

import pandas as pd

import gamspy._symbols as syms
import gamspy.utils as utils
from gamspy.exceptions import GamspyException, ValidationError

if TYPE_CHECKING:
    import io
    from pathlib import Path

    from gamspy import Container, Model, Options
    from gamspy._backend.engine import EngineClient, GAMSEngine
    from gamspy._backend.local import Local
    from gamspy._backend.neos import NeosClient, NEOSServer
    from gamspy._symbols.symbol import Symbol

SOLVE_STATUS = [
    "",
    "Normal",
    "Iteration",
    "Resource",
    "Solver",
    "EvalError",
    "Capability",
    "License",
    "User",
    "SetupErr",
    "SolverErr",
    "InternalErr",
    "Skipped",
    "SystemErr",
]
HEADER = [
    "Solver Status",
    "Model Status",
    "Objective",
    "Num of Equations",
    "Num of Variables",
    "Model Type",
    "Solver",
    "Solver Time",
]


@no_type_check
def backend_factory(
    container: Container,
    options: Options | None = None,
    solver: str | None = None,
    solver_options: dict | Path | None = None,
    output: io.TextIOWrapper | None = None,
    backend: Literal["local", "engine", "neos"] = "local",
    client: EngineClient | NeosClient | None = None,
    model: Model | None = None,
    load_symbols: list[Symbol] | None = None,
) -> Local | GAMSEngine | NEOSServer:
    if backend == "neos":
        from gamspy._backend.neos import NEOSServer

        return NEOSServer(
            container,
            options,
            solver,
            solver_options,
            client,
            output,
            model,
            load_symbols,
        )
    elif backend == "engine":
        from gamspy._backend.engine import GAMSEngine

        return GAMSEngine(
            container,
            client,
            options,
            solver,
            solver_options,
            output,
            model,
            load_symbols,
        )
    elif backend == "local":
        from gamspy._backend.local import Local

        return Local(
            container,
            options,
            solver,
            solver_options,
            output,
            model,
            load_symbols,
        )

    raise ValidationError(
        f"`{backend}` is not a valid backend. Possible backends:"
        " local, engine, and neos"
    )


def _cast_values(
    objective_value: str,
    num_equations: str,
    num_variables: str,
    solver_time: str,
) -> tuple[float, int | float, int | float, float]:
    objective = float("nan") if objective_value == "NA" else float(objective_value)
    equations = float("nan") if num_equations == "NA" else int(num_equations)
    variables = float("nan") if num_variables == "NA" else int(num_variables)
    time = float("nan") if solver_time == "NA" else float(solver_time)

    return objective, equations, variables, time


class Backend(ABC):
    def __init__(
        self,
        backend_type: str,
        container: Container,
        model: Model | None,
        options: Options,
        solver: str | None,
        solver_options: dict | Path | None,
        output: io.TextIOWrapper | None,
        load_symbols: list[Symbol] | None,
    ):
        self.backend_type = backend_type
        self.container = container
        self.model = model
        self.options = options
        self.solver = solver
        self.solver_options = solver_options
        self.output = output
        if load_symbols is not None:
            self.load_symbols: list[str] = [
                symbol.name  # type: ignore
                for symbol in load_symbols
            ]

        self.job_name = self.get_job_name()
        self.gms_file = self.job_name + ".gms"
        self.lst_file = self.job_name + ".lst"
        self.pf_file = self.job_name + ".pf"
        self.restart_file = self.job_name + ".g00"
        self.trace_file = self.job_name + ".txt"

    @abstractmethod
    def is_async(self): ...

    @abstractmethod
    def run(
        self,
        relaxed_domain_mapping: bool = False,
        gams_to_gamspy: bool = False,
    ): ...

    def get_job_name(self):
        job_name = self.container._job

        if self.container._debugging_level == "keep":
            job_name = os.path.join(
                self.container.working_directory,
                "_" + utils._get_unique_name(),
            )
            self.container._job = job_name
            self.container._gdx_in = f"{job_name}in.gdx"
            self.container._gdx_out = f"{job_name}out.gdx"

        return job_name

    def preprocess(self):
        modified_names = self.container._get_modified_symbols()

        if len(modified_names) != 0:
            try:
                self.container.write(
                    self.container._gdx_in, modified_names, eps_to_zero=False
                )
            except Exception as e:
                # Unfortunately, GTP raises a blind exception here. Turn it into a GamspyException.
                raise GamspyException(str(e)) from e

        gdx_in = self.container._gdx_in
        if self.backend_type == "engine":
            gdx_in = os.path.basename(gdx_in)
        elif self.backend_type == "neos":
            gdx_in = "in.gdx"
        gams_string = self.container._generate_gams_string(gdx_in, modified_names)
        self.make_unmodified(modified_names)

        return gams_string

    def load_records(self, relaxed_domain_mapping: bool = False):
        if hasattr(self, "load_symbols"):
            symbols = self.load_symbols
        else:
            symbols = utils._get_symbol_names_from_gdx(
                self.container.system_directory, self.container._gdx_out
            )
            filtered_names = []
            for name in symbols:
                # addGamsCode symbols
                if name not in self.container:
                    filtered_names.append(name)
                    continue

                symbol = self.container[name]
                if type(symbol) is syms.Alias:
                    filtered_names.append(name)
                    continue

                if symbol.synchronize:
                    filtered_names.append(name)

            symbols = filtered_names

        if len(symbols) != 0:
            self.container._load_records_from_gdx(self.container._gdx_out, symbols)
            self.make_unmodified(symbols)

        if relaxed_domain_mapping:
            # Best attempt approach to map relaxed domain to actual symbols
            for name in symbols:
                symbol = self.container[name]

                new_domain = []
                for elem in symbol.domain:
                    if type(elem) is str and elem != "*" and elem in self.container:
                        new_domain.append(self.container[elem])
                    else:
                        new_domain.append(elem)

                symbol.domain = new_domain
                symbol.dimension = len(new_domain)
                if type(symbol) in (syms.Variable, syms.Equation):
                    symbol._update_attr_domains()

    def parse_listings(self):
        listing_file = (
            self.options.listing_file
            if self.options.listing_file
            else self.job_name + ".lst"
        )
        if self.options.equation_listing_limit:
            utils._parse_generated_equations(self.model, listing_file)

        if self.options.variable_listing_limit:
            utils._parse_generated_variables(self.model, listing_file)

    def prepare_summary(self, trace_file: str) -> pd.DataFrame:
        from gamspy._model import ModelStatus

        with open(trace_file, encoding="utf-8") as file:
            line = file.readlines()[-1]
            (
                _,
                model_type,
                solver_name,
                _,
                _,
                _,
                _,
                num_equations,
                num_variables,
                _,
                _,
                _,
                _,
                model_status,
                solver_status,
                objective_value,
                _,
                solver_time,
                _,
                _,
                _,
                _,
            ) = line.split(",")

        objective_value, num_equations, num_variables, solver_time = _cast_values(
            objective_value, num_equations, num_variables, solver_time
        )

        dataframe = pd.DataFrame(
            [
                [
                    SOLVE_STATUS[int(solver_status)],
                    ModelStatus(int(model_status)).name,
                    objective_value,
                    num_equations,
                    num_variables,
                    model_type,
                    solver_name,
                    solver_time,
                ]
            ],
            columns=HEADER,
        )
        return dataframe

    def make_unmodified(self, modified_names: list[str]):
        for name in modified_names:
            if not name.startswith("autogenerated_"):
                self.container[name].modified = False



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

import gamspy._algebra.condition as condition
import gamspy._algebra.domain as domain
import gamspy._algebra.number as number
import gamspy._algebra.operable as operable
import gamspy._algebra.operation as operation
import gamspy._symbols as gp_syms
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy._config import get_option
from gamspy._extrinsic import ExtrinsicFunction
from gamspy._symbols.implicits import ImplicitSet
from gamspy._symbols.implicits.implicit_symbol import ImplicitSymbol
from gamspy._symbols.symbol import Symbol
from gamspy.exceptions import ValidationError
from gamspy.math.misc import MathOp

if TYPE_CHECKING:
    import pandas as pd

    from gamspy import Alias, Set
    from gamspy._symbols.implicits import ImplicitEquation
    from gamspy._types import OperableType

GMS_MAX_LINE_LENGTH = 80000
LINE_LENGTH_OFFSET = 79000


@dataclass
class DomainPlaceHolder:
    indices: list[tuple[str, int]]


def peek(stack):
    if len(stack) > 0:
        return stack[-1]
    return None


# Higher number means higher precedence.
PRECEDENCE = {
    "..": 0,
    "=": 0,
    "or": 1,
    "xor": 2,
    "and": 3,
    "=e=": 4,
    "=n=": 4,
    "=b=": 4,
    "eq": 4,
    "ne": 4,
    ">=": 4,
    "<=": 4,
    ">": 4,
    "<": 4,
    "=g=": 4,
    "=l=": 4,
    "=x=": 4,
    "+": 5,
    "-": 5,
    "*": 6,
    "/": 6,
    "not": 7,
    "u-": 7,
}

# Defines how operators of the same precedence are grouped.
ASSOCIATIVITY = {
    "or": "left",
    "xor": "left",
    "and": "left",
    "=e=": "left",
    "=n=": "left",
    "=x=": "left",
    "=b=": "left",
    "eq": "left",
    "ne": "left",
    "=": "left",
    ">=": "left",
    "<=": "left",
    ">": "left",
    "<": "left",
    "=g=": "left",
    "=l=": "left",
    "..": "non",
    "+": "left",
    "-": "left",
    "*": "left",
    "/": "left",
    "not": "right",
    "u-": "right",
}

# Precedence for a leaf node is considered infinite.
LEAF_PRECEDENCE = float("inf")


def get_operand_gams_repr(operand) -> str:
    if hasattr(operand, "gamsRepr"):
        return operand.gamsRepr()

    representation = str(operand)

    # b[i] * -1   -> not valid
    # b[i] * (-1) -> valid
    if isinstance(operand, (int, float)) and operand < 0:
        representation = f"({representation})"

    return representation


def get_operand_latex_repr(operand) -> str:
    if hasattr(operand, "latexRepr"):
        return operand.latexRepr()

    if isinstance(operand, float):
        operand = utils._map_special_values(operand)

    representation = str(operand)

    return representation


def create_gams_expression(root_node: Expression) -> str:
    """
    Creates GAMS representation of a binary expression tree without recursion.
    It uses an iterative post-order traversal to build the expression,
    adding parentheses only when necessary based on operator precedence and
    associativity rules.
    """
    if not isinstance(root_node, Expression):
        return get_operand_gams_repr(root_node)

    # 1. Get nodes in post-order (left - right - parent).
    s1: list[OperableType | ImplicitEquation | str] = [root_node]
    post_order_nodes = []
    while s1:
        node = s1.pop()
        post_order_nodes.append(node)
        if isinstance(node, Expression):
            if node.left is not None:
                s1.append(node.left)
            if node.right is not None:
                s1.append(node.right)

    # 2. Build the GAMS expression
    eval_stack: list[tuple[str, float]] = []
    for node in reversed(post_order_nodes):
        if not isinstance(node, Expression):
            eval_stack.append((get_operand_gams_repr(node), LEAF_PRECEDENCE))
            continue

        op = node.operator
        op_prec = PRECEDENCE[op]
        op_assoc = ASSOCIATIVITY[op]

        # Handle unary ops
        if op in ("u-", "not"):
            operand_str, operand_prec = eval_stack.pop()

            # Add parentheses if the operand's operator has lower precedence
            if operand_prec < op_prec:
                operand_str = f"({operand_str})"

            if op == "u-":
                new_str = f"(-{operand_str})"
                # A parenthesized expression has the highest precedence
                eval_stack.append((new_str, LEAF_PRECEDENCE))
            else:  # Standard handling for 'not'
                new_str = f"not {operand_str}"
                eval_stack.append((new_str, op_prec))

        # Handle binary ops
        else:
            right_str, right_prec = eval_stack.pop()
            left_str, left_prec = eval_stack.pop()

            if left_prec < op_prec or (left_prec == op_prec and op_assoc == "right"):
                left_str = f"({left_str})"

            if right_prec < op_prec or (right_prec == op_prec and op_assoc == "left"):
                right_str = f"({right_str})"

            # get around 80000 line length limitation in GAMS
            length = len(left_str) + len(op) + len(right_str)
            if length >= GMS_MAX_LINE_LENGTH - LINE_LENGTH_OFFSET:
                new_str = f"{left_str} {op}\n {right_str}"
            else:
                new_str = f"{left_str} {op} {right_str}"
            eval_stack.append((new_str, op_prec))

    final_string = eval_stack[0][0]

    if root_node.operator in ("=", ".."):
        return f"{final_string};"

    return final_string


def create_latex_expression(root_node: Expression) -> str:
    """
    Creates LaTeX representation of a binary expression tree without recursion.
    It uses an iterative post-order traversal to build the expression,
    adding parentheses only when necessary based on operator precedence and
    associativity rules.
    """
    if not isinstance(root_node, Expression):
        return get_operand_latex_repr(root_node)

    op_map = {
        "=g=": "\\geq",
        "=l=": "\\leq",
        "=e=": "=",
        "*": "\\cdot",
        "and": "\\wedge",
        "or": "\\vee",
        "xor": "\\oplus",
        "$": "|",
    }

    # 1. Get nodes in post-order (left - right - parent).
    s1: list[OperableType | ImplicitEquation | str] = [root_node]
    post_order_nodes = []
    while s1:
        node = s1.pop()
        post_order_nodes.append(node)
        if isinstance(node, Expression):
            if node.left is not None:
                s1.append(node.left)
            if node.right is not None:
                s1.append(node.right)

    # 2. Build the LaTeX expression
    eval_stack: list[tuple[str, float]] = []
    for node in reversed(post_order_nodes):
        if not isinstance(node, Expression):
            eval_stack.append((get_operand_latex_repr(node), LEAF_PRECEDENCE))
            continue

        op = node.operator
        op_prec = PRECEDENCE[op]
        op_assoc = ASSOCIATIVITY[op]

        # Handle unary ops
        if op in ("u-", "not"):
            operand_str, operand_prec = eval_stack.pop()

            # Add parentheses if the operand's operator has lower precedence
            if operand_prec < op_prec:
                operand_str = f"({operand_str})"

            if op == "u-":
                new_str = f"(-{operand_str})"
                # A parenthesized expression has the highest precedence
                eval_stack.append((new_str, LEAF_PRECEDENCE))
            else:  # Standard handling for 'not'
                new_str = f"not {operand_str}"
                eval_stack.append((new_str, op_prec))

        # Handle binary ops
        else:
            right_str, right_prec = eval_stack.pop()
            left_str, left_prec = eval_stack.pop()

            if left_prec < op_prec or (left_prec == op_prec and op_assoc == "right"):
                left_str = f"({left_str})"

            if right_prec < op_prec or (right_prec == op_prec and op_assoc == "left"):
                right_str = f"({right_str})"

            if op == "/":
                eval_stack.append((f"\\frac{{{left_str}}}{{{right_str}}}", op_prec))
                continue

            op = op_map.get(op, op)

            # get around 80000 line length limitation in GAMS
            length = len(left_str) + len(op) + len(right_str)
            if length >= GMS_MAX_LINE_LENGTH - LINE_LENGTH_OFFSET:
                new_str = f"{left_str} {op}\n {right_str}"
            else:
                new_str = f"{left_str} {op} {right_str}"
            eval_stack.append((new_str, op_prec))

    final_string = eval_stack[0][0]

    return final_string


class Expression(operable.Operable):
    """
    Expression of two operands and an operation.

    Parameters
    ----------
    left: str | int | float | Parameter | Variable | None
        Left operand
    data: str
        Operation
    right: str | int | float | Parameter | Variable | None
        Right operand

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> a = gp.Parameter(m, name="a")
    >>> b = gp.Parameter(m, name="b")
    >>> expression = a * b
    >>> expression.gamsRepr()
    'a * b'

    """

    def __init__(
        self,
        left: OperableType | ImplicitEquation | None,
        operator: str,
        right: OperableType | str | None,
    ):
        self.left = utils._map_special_values(left) if isinstance(left, float) else left
        self.operator = operator
        self.right = (
            utils._map_special_values(right) if isinstance(right, float) else right
        )

        if operator == "=" and isinstance(right, Expression):
            right._fix_equalities()

        self._representation: str | None = None
        self.where = condition.Condition(self)
        self._create_domain()
        left_control = getattr(left, "controlled_domain", [])
        right_control = getattr(right, "controlled_domain", [])
        self.controlled_domain: list[Set | Alias] = list(
            {*left_control, *right_control}
        )
        self.container = None
        if hasattr(left, "container"):
            self.container = left.container  # type: ignore
        elif hasattr(right, "container"):
            self.container = right.container  # type: ignore

    @property
    def representation(self) -> str:
        if self._representation is None:
            self._representation = create_gams_expression(self)

        return self._representation

    def _create_domain(self):
        for loc, result in (
            (self.left, "_left_domain"),
            (self.right, "_right_domain"),
        ):
            if isinstance(loc, condition.Condition):
                loc = loc.conditioning_on

            if loc is None or isinstance(loc, (int, float, str)):
                result_domain = []  # left is a scalar
            elif isinstance(loc, domain.Domain):
                result_domain = loc.sets
            else:
                result_domain = loc.domain

            setattr(self, result, result_domain)

        left_domain = self._left_domain
        right_domain = self._right_domain

        set_to_index = {}

        for domain_char, domain_ptr in (
            ("l", left_domain),
            ("r", right_domain),
        ):
            for i, d in enumerate(domain_ptr):
                if isinstance(d, str):
                    continue  # string domains are fixed and they do not count

                if d not in set_to_index:
                    set_to_index[d] = []

                set_to_index[d].append((domain_char, i))

        shadow_domain = []
        result_domain = []
        for d in (*left_domain, *right_domain):
            if isinstance(d, str):
                continue

            if d not in result_domain:
                result_domain.append(d)
                indices = set_to_index[d]
                shadow_domain.append(DomainPlaceHolder(indices=indices))

        self._shadow_domain = shadow_domain
        self.domain = result_domain
        self.dimension = validation.get_dimension(self.domain)

    def __getitem__(self, indices):
        indices = validation.validate_domain(self, indices)
        left_domain = list(self._left_domain)
        right_domain = list(self._right_domain)
        for i, s in enumerate(indices):
            for lr, pos in self._shadow_domain[i].indices:
                if lr == "l":
                    left_domain[pos] = s
                else:
                    right_domain[pos] = s

        left = self.left[left_domain] if left_domain else self.left
        right = self.right[right_domain] if right_domain else self.right

        return Expression(left, self.operator, right)

    @property
    def records(self) -> pd.DataFrame | None:
        """
        Evaluates the expression and returns the resulting records.

        Returns
        -------
        pd.DataFrame | None

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> a = gp.Parameter(m, records=5)
        >>> b = gp.Parameter(m, records=6)
        >>> (a + b).records
           value
        0   11.0

        """
        assert self.container is not None
        temp_name = "a" + utils._get_unique_name()
        temp_param = gp_syms.Parameter._constructor_bypass(
            self.container, temp_name, self.domain
        )
        temp_param[...] = self
        del self.container.data[temp_name]
        return temp_param.records

    def toValue(self) -> float | None:
        """
        Convenience method to return expression records as a Python float. Only possible if there is a single record as a result of the expression evaluation.

        Returns
        -------
        float | None

        Raises
        ------
        TypeError
            In case the dimension of the expression is not zero.

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> a = gp.Parameter(m, records=5)
        >>> b = gp.Parameter(m, records=6)
        >>> (a + b).toValue()
        np.float64(11.0)

        """
        if self.dimension != 0:
            raise TypeError(
                f"Cannot extract value data for non-scalar expressions (expression dimension is {self.dimension})"
            )

        records = self.records
        if records is not None:
            return records["value"][0]

        return records

    def toList(self) -> list | None:
        """
        Convenience method to return the records of the expression as a list.

        Returns
        -------
        list | None

        Examples
        --------
        >>> import numpy as np
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, records=range(3))
        >>> a = gp.Parameter(m, domain=i, records=np.array([1,2,3]))
        >>> b = gp.Parameter(m, domain=i, records=np.array([4,5,6]))
        >>> (a + b).toList()
        [['0', 5.0], ['1', 7.0], ['2', 9.0]]

        """
        records = self.records
        if records is not None:
            return records.values.tolist()

        return None

    def __eq__(self, other):
        return Expression(self, "=e=", other)

    def __ne__(self, other):
        return Expression(self, "ne", other)

    def __bool__(self):
        raise ValidationError(
            "An expression cannot be used as a truth value. If you are "
            "trying to generate an expression, use binary operators "
            "instead (e.g. &, |, ^). For more details, see: "
            "https://gamspy.readthedocs.io/en/latest/user/gamspy_for_gams_users.html#logical-operations"
        )

    def __repr__(self) -> str:
        return f"Expression(left={self.left}, data={self.operator}, right={self.right})"

    def _replace_operator(self, operator: str):
        self.operator = operator

    def latexRepr(self) -> str:
        """
        Representation of this Expression in Latex.

        Returns
        -------
        str
        """
        return create_latex_expression(self)

    def gamsRepr(self) -> str:
        """
        Representation of this Expression in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> a = gp.Parameter(m, name="a")
        >>> b = gp.Parameter(m, name="b")
        >>> expression = a * b
        >>> expression.gamsRepr()
        'a * b'

        """
        return self.representation

    def getDeclaration(self) -> str:
        """
        Declaration of the Expression in GAMS

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> a = gp.Parameter(m, name="a")
        >>> b = gp.Parameter(m, name="b")
        >>> expression = a * b
        >>> expression.getDeclaration()
        'a * b'

        """
        return self.gamsRepr()

    def _fix_equalities(self) -> None:
        # Equality operations on Parameter and Variable objects generate
        # GAMS equality signs: =g=, =e=, =l=. If these signs appear on
        # assignments, replace them with regular equality ops.
        # Uses a stack based post-order traversal algorithm.
        EQ_MAP: dict[str, str] = {"=g=": ">=", "=e=": "eq", "=l=": "<="}
        stack = []
        root = self

        while True:
            while root is not None:
                if hasattr(root, "right"):
                    stack.append(root.right)

                stack.append(root)
                root = root.left if hasattr(root, "left") else None  # type: ignore

            if len(stack) == 0:
                break

            root = stack.pop()

            if isinstance(root, Expression) and root.operator in EQ_MAP:
                root._replace_operator(EQ_MAP[root.operator])

            last_item = peek(stack)
            if (
                hasattr(root, "right")
                and last_item is not None
                and last_item is root.right
            ):
                stack.pop()
                stack.append(root)
                root = root.right
            else:
                root = None

    def _find_all_symbols(self) -> list[str]:
        # Finds all symbols in an expression with a stack based inorder
        # traversal algorithm (O(N)).
        symbols: list[str] = []
        stack = []

        node = self
        while True:
            if node is not None:
                stack.append(node)
                node = getattr(node, "left", None)  # type: ignore
            elif stack:
                node = stack.pop()

                if isinstance(node, Symbol):
                    if node.name not in symbols:
                        if type(node) is gp_syms.Alias:
                            symbols.append(node.alias_with.name)

                        symbols.append(node.name)
                    stack.extend(node.domain)
                    node = None
                elif isinstance(node, ImplicitSymbol):
                    if node.parent.name not in symbols:
                        symbols.append(node.parent.name)
                    stack.extend(node.domain)
                    stack.extend(node.container[node.parent.name].domain)
                    node = None
                elif isinstance(node, operation.Operation):
                    stack.extend(node.op_domain)
                    node = node.rhs
                elif isinstance(node, condition.Condition):
                    stack.append(node.conditioning_on)

                    if isinstance(node.condition, Expression):
                        node = node.condition
                    else:
                        stack.append(node.condition)
                        node = None
                elif isinstance(node, (operation.Ord, operation.Card)):
                    stack.append(node._symbol)
                    node = None
                elif isinstance(node, MathOp):
                    if isinstance(node.elements[0], Expression):
                        node = node.elements[0]
                    else:
                        stack.extend(node.elements)
                        node = None
                elif isinstance(node, ExtrinsicFunction):
                    stack.extend(list(node.args))
                    node = None
                else:
                    node = getattr(node, "right", None)
            else:
                break  # pragma: no cover

        return symbols

    def _find_symbols_in_conditions(self) -> list[str]:
        symbols: list[str] = []
        stack = []

        node = self
        while True:
            if node is not None:
                stack.append(node)
                node = getattr(node, "left", None)  # type: ignore
            elif stack:
                node = stack.pop()

                if isinstance(node, condition.Condition):
                    given_condition = node.condition

                    if isinstance(given_condition, Expression):
                        symbols.extend(given_condition._find_all_symbols())
                    elif isinstance(given_condition, ImplicitSymbol):
                        symbols.append(given_condition.parent.name)

                if isinstance(node, operation.Operation):
                    stack.extend(node.op_domain)
                    node = node.rhs
                else:
                    node = getattr(node, "right", None)
            else:
                break  # pragma: no cover

        return symbols

    def _validate_definition(
        self, control_stack: list[Set | Alias | ImplicitSet]
    ) -> None:
        if not get_option("DOMAIN_VALIDATION") or not get_option("DOMAIN_VALIDATION"):
            return

        stack = []

        node = self.right
        while True:
            if node is not None:
                stack.append(node)
                node = getattr(node, "left", None)  # type: ignore
            elif stack:
                node = stack.pop()

                if isinstance(node, operation.Operation):
                    node._validate_operation(control_stack.copy())
                elif isinstance(node, ImplicitSymbol):
                    for elem in node.domain:
                        if hasattr(elem, "is_singleton") and elem.is_singleton:
                            continue

                        if isinstance(elem, Symbol) and elem not in control_stack:
                            raise ValidationError(
                                f"Uncontrolled set `{elem}` entered as constant!"
                            )
                        elif (
                            isinstance(elem, ImplicitSymbol)
                            and elem.parent not in control_stack
                        ):
                            raise ValidationError(
                                f"Uncontrolled set `{elem.parent}` entered as constant!"
                            )

                node = getattr(node, "right", None)
            else:
                break  # pragma: no cover


class SetExpression(Expression):
    def __init__(
        self,
        left: OperableType,
        data: Literal["+", "-", "*", "not"],
        right: OperableType,
    ):
        super().__init__(left, data, right)
        self._adjust_left_right()

    def _adjust_left_right(self) -> None:
        if isinstance(self.left, (ImplicitSet, SetExpression)):
            if isinstance(self.right, (int, float)):
                if self.right == 0:
                    self.right = "no"
                elif self.right == 1:
                    self.right = "yes"
                else:
                    raise ValidationError(
                        f"Incompatible operand `{self.right}` for the set operation `{self.operator}`."
                    )
            elif isinstance(self.right, condition.Condition) and isinstance(
                self.right.conditioning_on, number.Number
            ):
                if self.right.conditioning_on._value == 0:
                    self.right.conditioning_on._value = "no"
                elif self.right.conditioning_on._value == 1:
                    self.right.conditioning_on._value = "yes"
                raise ValidationError(
                    f"Incompatible operand `{self.right}` for the set operation `{self.operator}`."
                )

        if isinstance(self.right, (ImplicitSet, SetExpression)):
            if isinstance(self.left, (int, float)):
                if self.left == 0:
                    self.left = "no"
                elif self.left == 1:
                    self.left = "yes"
                else:
                    raise ValidationError(
                        f"Incompatible operand `{self.left}` for the set operation `{self.operator}`."
                    )
            elif isinstance(self.left, condition.Condition) and isinstance(
                self.left.conditioning_on, number.Number
            ):
                if self.left.conditioning_on._value == 0:
                    self.left.conditioning_on._value = "no"
                elif self.left.conditioning_on._value == 1:
                    self.left.conditioning_on._value = "yes"
                else:
                    raise ValidationError(
                        f"Incompatible operand `{self.left}` for the set operation `{self.operator}`."
                    )



from __future__ import annotations

from typing import Literal

import gamspy as gp
import gamspy.formulations.utils as utils
from gamspy.exceptions import ValidationError
from gamspy.math import dim


class _MPool2d:
    def __init__(
        self,
        sense: Literal["min", "max"],
        container: gp.Container,
        kernel_size: int | tuple[int, int],
        stride: int | tuple[int, int] | None = None,
        padding: int = 0,
        name_prefix: str | None = None,
    ):
        # Validate pooling type
        if sense not in ("min", "max"):
            raise ValidationError("_MPool2d expects min or max")

        # Convert kernel_size, stride, and padding to tuples
        _kernel_size = utils._check_tuple_int(kernel_size, "kernel_size")
        if stride is None:
            stride = kernel_size  # Default stride = kernel size

        _stride = utils._check_tuple_int(stride, "stride")
        _padding = utils._check_tuple_int(padding, "padding", allow_zero=True)

        # Store configurations
        self.container = container
        self.kernel_size = _kernel_size
        self.stride = _stride
        self.padding = _padding
        self.sense = sense

        if name_prefix is None:
            name_prefix = gp.utils._get_unique_name()

        self._name_prefix = name_prefix

    def _set_bounds_and_big_M(
        self,
        input: gp.Parameter | gp.Variable,
        default_big_m: int,
        subset: gp.Set,
        out_var: gp.Variable,
    ) -> gp.Parameter:
        # Extract batch and channel dimensions from input domain
        N, C = input.domain[:2]
        H_out, W_out, Hf, Wf, H_in, W_in = subset.domain

        # Create subset2 mapping output positions to input positions
        subset2 = gp.Set(
            self.container,
            name=utils._generate_name("s", self._name_prefix, "in_out_matching_2"),
            domain=[H_out, W_out, H_in, W_in],
        )
        subset2[H_out, W_out, H_in, W_in] = gp.Sum(
            [Hf, Wf], subset[H_out, W_out, Hf, Wf, H_in, W_in]
        )

        # Initialize parameters for bounds and Big-M
        big_m_par = gp.Parameter(
            self.container,
            name=utils._generate_name("p", self._name_prefix, "bigM_1"),
            domain=[N, C, H_out, W_out],
        )
        lb = gp.Parameter(
            self.container,
            name=utils._generate_name("p", self._name_prefix, "output_lb"),
            domain=[N, C, H_out, W_out],
        )
        ub = gp.Parameter(
            self.container,
            name=utils._generate_name("p", self._name_prefix, "output_ub"),
            domain=[N, C, H_out, W_out],
        )

        # Calculate upper/lower bounds based on input type
        if isinstance(input, gp.Variable):
            # Use variable bounds if input is a variable
            ub[...] = gp.Smax(
                gp.Domain(H_in, W_in).where[subset2[H_out, W_out, H_in, W_in]],
                input.up[N, C, H_in, W_in],
            )
            lb[...] = gp.Smin(
                gp.Domain(H_in, W_in).where[subset2[H_out, W_out, H_in, W_in]],
                input.lo[N, C, H_in, W_in],
            )
        else:
            # Use parameter values directly if input is a parameter
            ub[...] = gp.Smax(
                gp.Domain(H_in, W_in).where[subset2[H_out, W_out, H_in, W_in]],
                input[N, C, H_in, W_in],
            )
            lb[...] = gp.Smin(
                gp.Domain(H_in, W_in).where[subset2[H_out, W_out, H_in, W_in]],
                input[N, C, H_in, W_in],
            )

        # Set output variable bounds
        out_var.lo[...] = lb
        out_var.up[...] = ub

        # Calculate Big-M values and handle infinity
        big_m_par[N, C, H_out, W_out] = ub - lb
        big_m_par[...].where[big_m_par[...] == "inf"] = default_big_m

        return big_m_par

    def __call__(
        self,
        input: gp.Parameter | gp.Variable,
        big_m: int = 1000,
        propagate_bounds: bool = True,
    ) -> tuple[gp.Variable, list[gp.Equation]]:
        # User input validation
        if not isinstance(input, (gp.Parameter, gp.Variable)):
            raise ValidationError("Expected a parameter or a variable input")

        if not isinstance(propagate_bounds, bool):
            raise ValidationError("Expected a boolean for propagate_bounds")

        if len(input.domain) != 4:
            raise ValidationError(f"expected 4D input (got {len(input.domain)}D input)")

        # Extract dimensions from input (Batch, Channel, Height, Width)
        N, C, H_in, W_in = input.domain
        h_in = len(H_in)
        w_in = len(W_in)

        # Calculate output dimensions using padding, kernel, and stride
        h_out, w_out = utils._calc_hw(
            self.padding, self.kernel_size, self.stride, h_in, w_in
        )

        # Create output variable
        out_var = gp.Variable(
            self.container,
            name=utils._generate_name("v", self._name_prefix, "output"),
            domain=dim([len(N), len(C), h_out, w_out]),
        )
        N, C, H_out, W_out = out_var.domain

        # Calculate input window positions (top - left)
        # These indices determine where pooling windows start in the input tensor
        # Formula accounts for padding and stride
        top_index = (self.stride[0] * (gp.Ord(H_out) - 1)) - self.padding[0] + 1
        left_index = (self.stride[1] * (gp.Ord(W_out) - 1)) - self.padding[1] + 1

        # Create filter dimensions and domain relationships
        Hf, Wf = gp.math._generate_dims(self.container, self.kernel_size)
        Hf, Wf, H_in, W_in = utils._next_domains([Hf, Wf, H_in, W_in], out_var.domain)

        # Create mapping between input/output positions
        subset = gp.Set(
            self.container,
            name=utils._generate_name("s", self._name_prefix, "in_out_matching_1"),
            domain=[H_out, W_out, Hf, Wf, H_in, W_in],
        )
        # Create relationship between output positions and their corresponding input windows
        subset[
            H_out,
            W_out,
            Hf,
            Wf,
            H_in,
            W_in,
        ].where[
            (gp.Ord(H_in) == (top_index + gp.Ord(Hf) - 1))  # Vertical position
            & (gp.Ord(W_in) == (left_index + gp.Ord(Wf) - 1))  # Horizontal
        ] = True

        # Create a binary variable
        bin_var = gp.Variable(
            self.container,
            name=utils._generate_name("v", self._name_prefix, "aux_variable"),
            type="binary",
            domain=[N, C, H_out, W_out, H_in, W_in],
        )

        # Create constraint equations
        greater_than = gp.Equation(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "gte"),
            domain=[N, C, H_out, W_out, Hf, Wf, H_in, W_in],
        )
        less_than = gp.Equation(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "lte"),
            domain=[N, C, H_out, W_out, Hf, Wf, H_in, W_in],
        )

        # Set up Big-M parameter
        if propagate_bounds:
            _big_m = self._set_bounds_and_big_M(input, big_m, subset, out_var)
        else:
            _big_m = gp.Parameter(
                self.container,
                name=utils._generate_name("p", self._name_prefix, "bigM_2"),
                records=big_m,
            )

        big_m_expr = _big_m * (1 - bin_var[N, C, H_out, W_out, H_in, W_in])

        # Build constraints based on pooling type
        if self.sense == "max":
            # For max pooling:
            greater_than[N, C, subset[H_out, W_out, Hf, Wf, H_in, W_in]] = (
                out_var[N, C, H_out, W_out] >= input[N, C, H_in, W_in]
            )
            less_than[N, C, subset[H_out, W_out, Hf, Wf, H_in, W_in]] = (
                out_var[N, C, H_out, W_out] <= input[N, C, H_in, W_in] + big_m_expr
            )
        else:
            # For min pooling:
            greater_than[N, C, subset[H_out, W_out, Hf, Wf, H_in, W_in]] = (
                out_var[N, C, H_out, W_out] + big_m_expr >= input[N, C, H_in, W_in]
            )
            less_than[N, C, subset[H_out, W_out, Hf, Wf, H_in, W_in]] = (
                out_var[N, C, H_out, W_out] <= input[N, C, H_in, W_in]
            )

        # Ensure exactly one element is selected per window
        pick_one = gp.Equation(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "pick_one"),
            domain=[N, C, H_out, W_out],
        )
        pick_one[N, C, H_out, W_out] = (
            gp.Sum(
                [Hf, Wf],
                gp.Sum(
                    subset[H_out, W_out, Hf, Wf, H_in, W_in],
                    bin_var[N, C, H_out, W_out, H_in, W_in],
                ),
            )
            == 1
        )

        return out_var, [less_than, greater_than, pick_one]



from __future__ import annotations

import os
from typing import Any, Literal

from gamspy.exceptions import ValidationError

configuration: dict[str, Any] = {}


def _set_default_options() -> None:
    # Set the default for GAMS_SYSDIR
    try:
        import gamspy_base

        configuration["GAMS_SYSDIR"] = gamspy_base.directory
    except ModuleNotFoundError:
        ...

    sysdir = os.getenv("GAMSPY_GAMS_SYSDIR", None)
    if sysdir is not None:
        configuration["GAMS_SYSDIR"] = sysdir

    # Enable all validations by default
    validate = int(os.getenv("GAMSPY_VALIDATION", 1))
    configuration["VALIDATION"] = validate

    # Check for domain violation by default
    validate = int(os.getenv("GAMSPY_DOMAIN_VALIDATION", 1))
    configuration["DOMAIN_VALIDATION"] = validate

    # Solver validation. Enabled by default.
    validate_solver = int(os.getenv("GAMSPY_SOLVER_VALIDATION", 1))
    configuration["SOLVER_VALIDATION"] = validate_solver

    # Solver option validation. Enabled by default.
    validate_solver_options = int(os.getenv("GAMSPY_SOLVER_OPTION_VALIDATION", 1))
    configuration["SOLVER_OPTION_VALIDATION"] = validate_solver_options

    # Special value mapping
    map_special_values = int(os.getenv("GAMSPY_MAP_SPECIAL_VALUES", 1))
    configuration["MAP_SPECIAL_VALUES"] = map_special_values

    # Assume .l or .scale for variables in assignments
    assume_level = int(os.getenv("GAMSPY_ASSUME_VARIABLE_SUFFIX", 1))
    configuration["ASSUME_VARIABLE_SUFFIX"] = assume_level

    # Try to use the Python variable name in case the name is not provided.
    use_py_var_name = os.getenv("GAMSPY_USE_PY_VAR_NAME", "yes-or-autogenerate")
    configuration["USE_PY_VAR_NAME"] = use_py_var_name

    # Do not drop the domain violations by default
    strategy = int(os.getenv("GAMSPY_DROP_DOMAIN_VIOLATIONS", 0))
    configuration["DROP_DOMAIN_VIOLATIONS"] = strategy

    # Do not allow ambiguity in MCP, EMP, MPEC models by default
    ambiguity = os.getenv("GAMSPY_ALLOW_AMBIGUOUS_EQUATIONS", "auto")
    configuration["ALLOW_AMBIGUOUS_EQUATIONS"] = ambiguity


def set_options(
    options: dict[
        Literal[
            "GAMS_SYSDIR",
            "VALIDATION",
            "DOMAIN_VALIDATION",
            "SOLVER_OPTION_VALIDATION",
            "MAP_SPECIAL_VALUES",
            "ASSUME_VARIABLE_SUFFIX",
            "USE_PY_VAR_NAME",
            "DROP_DOMAIN_VIOLATIONS",
            "ALLOW_AMBIGUOUS_EQUATIONS",
        ],
        Any,
    ],
) -> None:
    """
    Sets the given configuration options.

    Parameters
    ----------
    options : dict[str, Any]

    Raises
    ------
    ValidationError
        In case the given options are not in dict type.

    Examples
    --------
    >>> import gamspy as gp
    >>> gp.set_options({"DOMAIN_VALIDATION": 1})

    """
    if not isinstance(options, dict):
        raise ValidationError(
            f"`options` must be a dictionary but found: `{type(options)}`"
        )

    for key, value in options.items():
        configuration[key] = value


def get_option(name: str) -> Any:
    """
    Returns the requested option.

    Parameters
    ----------
    name : str
        Option name.

    Returns
    -------
    Any
        The value of the option.

    Raises
    ------
    KeyError
        In case the option is not set.

    Examples
    --------
    >>> import gamspy as gp
    >>> gp.set_options({"DOMAIN_VALIDATION": 0})
    >>> gp.get_option("DOMAIN_VALIDATION")
    0
    >>> gp.set_options({"DOMAIN_VALIDATION": 1})

    """
    return configuration[name]



from __future__ import annotations

import os
import platform
import signal
import socket
import subprocess
import threading
import time
from typing import TYPE_CHECKING

import certifi

from gamspy import utils
from gamspy.exceptions import FatalError, GamspyException, ValidationError

if TYPE_CHECKING:
    import io

    from gamspy import Container


_comm_pairs: dict[str, tuple[socket.socket, subprocess.Popen]] = {}


def open_connection(container: Container) -> None:
    LOOPBACK = "127.0.0.1"
    TIMEOUT = 30

    initial_pf_file = os.path.join(container._process_directory, "gamspy.pf")
    with open(initial_pf_file, "w") as file:
        file.write(
            'incrementalMode="2"\n'
            f'procdir="{container._process_directory}"\n'
            f'license="{container._license_path}"\n'
            f'curdir="{os.getcwd()}"\n'
        )

    command = [
        os.path.join(container.system_directory, "gams"),
        "GAMSPY_JOB",
        "pf",
        initial_pf_file,
    ]

    certificate_path = os.path.join(utils.DEFAULT_DIR, "gamspy_cert.crt")
    env = os.environ.copy()
    if os.path.isfile(certificate_path):
        env["GAMSLICECRT"] = certificate_path

    env = os.environ.copy()
    if "CURL_CA_BUNDLE" not in env:
        env["CURL_CA_BUNDLE"] = certifi.where()

    process = subprocess.Popen(
        command,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        errors="replace",
        start_new_session=platform.system() != "Windows",
        env=env,
    )

    port_info = process.stdout.readline().strip()

    try:
        port = int(port_info.removeprefix("port: "))
    except ValueError as e:
        raise ValidationError(
            f"Error while reading the port! {port_info + process.stdout.read()}"
        ) from e

    def handler(signum, frame):
        if platform.system() != "Windows":
            os.kill(process.pid, signal.SIGINT)

    if threading.current_thread() is threading.main_thread():
        signal.signal(signal.SIGINT, handler)

    start = time.time()
    while True:
        if process.poll() is not None:  # pragma: no cover
            raise ValidationError(process.communicate()[0])

        try:
            new_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            new_socket.connect((LOOPBACK, port))
            break
        except (ConnectionRefusedError, OSError) as e:
            new_socket.close()
            end = time.time()

            if end - start > TIMEOUT:  # pragma: no cover
                raise FatalError(
                    f"Timeout while establishing the connection with socket. {process.communicate()[0]}"
                ) from e

    _comm_pairs[container._comm_pair_id] = (new_socket, process)


def get_connection(pair_id: str) -> tuple[socket.socket, subprocess.Popen]:
    return _comm_pairs[pair_id]


def close_connection(pair_id: str):
    try:
        _socket, process = get_connection(pair_id)
    except KeyError:
        # This means that the connection is already closed.
        return

    _socket.sendall(b"stop")
    _socket.close()

    if not process.stdout.closed:
        process.stdout.close()

    # Wait until the GAMS process dies.
    while process.poll() is None:
        ...

    del _comm_pairs[pair_id]


def _read_output(process: subprocess.Popen, output: io.TextIOWrapper | None) -> None:
    if output is not None:
        while True:
            data = process.stdout.readline()
            output.write(data)
            output.flush()
            if data.startswith("--- Job ") and "elapsed" in data:
                break


def check_response(response: bytes, job_name: str) -> None:
    GAMS_STATUS = {
        1: "Solver is to be called, the system should never return this number.",
        2: "There was a compilation error.",
        3: "There was an execution error.",
        4: "System limits were reached.",
        5: "There was a file error.",
        6: "There was a parameter error.",
        7: "The solve has failed due to a license error. The license you are using may impose model size limits (demo/community license) or you are using a GAMSPy incompatible professional license. Please contact sales@gams.com to find out about license options.",
        8: "There was a GAMS system error.",
        9: "GAMS could not be started.",
        10: "Out of memory.",
        11: "Out of disk.",
        109: "Could not create process/scratch directory.",
        110: "Too many process/scratch directories.",
        112: "Could not delete the process/scratch directory.",
        113: "Could not write the script gamsnext.",
        114: "Could not write the parameter file.",
        115: "Could not read environment variable.",
        400: "Could not spawn the GAMS language compiler (gamscmex).",
        401: "Current directory (curdir) does not exist.",
        402: "Cannot set current directory (curdir).",
        404: "Blank in system directory (UNIX only).",
        405: "Blank in current directory (UNIX only).",
        406: "Blank in scratch extension (scrext)",
        407: "Unexpected cmexRC.",
        408: "Could not find the process directory (procdir).",
        409: "CMEX library not be found (experimental).",
        410: "Entry point in CMEX library could not be found (experimental).",
        411: "Blank in process directory (UNIX only).",
        412: "Blank in scratch directory (UNIX only).",
        909: "Cannot add path / unknown UNIX environment / cannot set environment variable.",
        1000: "Driver error: incorrect command line parameters for gams.",
        2000: "Driver error: internal error: cannot install interrupt handler.",
        3000: "Driver error: problems getting current directory.",
        4000: "Driver error: internal error: GAMS compile and execute module not found.",
        5000: "Driver error: internal error: cannot load option handling library.",
    }

    value = response[: response.find(b"#")].decode("ascii")
    if not value:
        raise FatalError(
            "Error while getting the return code from GAMS backend. This means that GAMS is in a bad state. Try to backtrack for previous errors."
        )

    return_code = int(value)

    if return_code in GAMS_STATUS:
        try:
            info = GAMS_STATUS[return_code]
        except IndexError:  # pragma: no cover
            info = ""

        raise GamspyException(
            f"Return code {return_code}. {info} Check {job_name + '.lst'} for more information.",
            return_code,
        )


def send_job(
    comm_pair_id: str,
    job_name: str,
    pf_file: str,
    output: io.TextIOWrapper | None = None,
):
    _socket, process = get_connection(comm_pair_id)
    try:
        # Send pf file
        _socket.sendall(pf_file.encode("utf-8"))

        # Read output
        _read_output(process, output)

        # Receive response
        response = _socket.recv(256)
        check_response(response, job_name)
    except ConnectionError as e:
        raise FatalError(
            f"There was an error while communicating with GAMS server: {e}",
        ) from e



from __future__ import annotations

import math
import typing

import gamspy._algebra.expression as expression
import gamspy.math as gamspy_math
from gamspy.exceptions import ValidationError

if typing.TYPE_CHECKING:
    from gamspy._types import OperableType


class Operable:
    """
    A mixin class that overloads the magic operations of a class
    to be used in Expressions
    """

    def __iter__(self):
        raise ValidationError(
            "GAMSPy symbols are not iterable. If you want to iterate on records, iterate over <symbol>.records."
        )

    # +, -, /, *, **, %
    def __add__(self, other: OperableType):
        from gamspy._symbols import Alias, Set
        from gamspy._symbols.implicits import ImplicitSet

        if isinstance(self, (Alias, Set)) and not isinstance(
            other, (Alias, Set, ImplicitSet)
        ):
            return self.lead(other)

        if isinstance(self, (ImplicitSet, expression.SetExpression)) or isinstance(
            other, (ImplicitSet, expression.SetExpression)
        ):
            return expression.SetExpression(self, "+", other)

        return expression.Expression(self, "+", other)

    def __radd__(self, other: OperableType):
        from gamspy._symbols.implicits import ImplicitSet

        if isinstance(self, (ImplicitSet, expression.SetExpression)) or isinstance(
            other, (ImplicitSet, expression.SetExpression)
        ):
            return expression.SetExpression(other, "+", self)

        return expression.Expression(other, "+", self)

    def __sub__(self, other: OperableType):
        from gamspy._symbols import Alias, Set
        from gamspy._symbols.implicits import ImplicitSet

        if isinstance(self, (Alias, Set)) and not isinstance(
            other, (Alias, Set, ImplicitSet)
        ):
            return self.lag(other)

        if isinstance(self, (ImplicitSet, expression.SetExpression)) or isinstance(
            other, (ImplicitSet, expression.SetExpression)
        ):
            return expression.SetExpression(self, "-", other)

        return expression.Expression(self, "-", other)

    def __rsub__(self, other: OperableType):
        from gamspy._symbols.implicits import ImplicitSet

        if isinstance(self, (ImplicitSet, expression.SetExpression)) or isinstance(
            other, (ImplicitSet, expression.SetExpression)
        ):
            return expression.SetExpression(other, "-", self)

        return expression.Expression(other, "-", self)

    def __neg__(self):
        return expression.Expression(None, "u-", self)

    def __truediv__(self, other: OperableType):
        return expression.Expression(self, "/", other)

    def __rtruediv__(self, other: OperableType):
        return expression.Expression(other, "/", self)

    def __mul__(self, other: OperableType):
        from gamspy._symbols.implicits import ImplicitSet

        if isinstance(self, (ImplicitSet, expression.SetExpression)) or isinstance(
            other, (ImplicitSet, expression.SetExpression)
        ):
            return expression.SetExpression(self, "*", other)

        return expression.Expression(self, "*", other)

    def __rmul__(self, other: OperableType):
        from gamspy._symbols.implicits import ImplicitSet

        if isinstance(self, (ImplicitSet, expression.SetExpression)) or isinstance(
            other, (ImplicitSet, expression.SetExpression)
        ):
            return expression.SetExpression(other, "*", self)

        return expression.Expression(other, "*", self)

    @typing.no_type_check
    def __pow__(self, other: OperableType):
        if (
            isinstance(other, int)
            and other == 2
            and isinstance(self, expression.Expression)
            and isinstance(self.left, gamspy_math.misc.MathOp)
            and self.left.op_name == "sqrt"
            and self.left.safe_cancel
        ):
            return self.left.elements[0]

        if isinstance(other, int):
            return gamspy_math.power(self, other)
        elif isinstance(other, float):
            if other == 0.5:
                return gamspy_math.sqrt(self)
            elif math.isclose(other, round(other), rel_tol=1e-4):
                return gamspy_math.power(self, other)

        return gamspy_math.rpower(self, other)

    def __rpow__(self, other: int | float):
        # e.g. 2 ** a[i] -> where 2 is other and a[i] is self.
        return gamspy_math.rpower(other, self)

    def __mod__(self, other: OperableType):
        return gamspy_math.mod(self, other)

    # and, or, xor
    def __and__(self, other: OperableType):
        return expression.Expression(self, "and", other)

    def __rand__(self, other: OperableType):
        return expression.Expression(other, "and", self)

    def __or__(self, other: OperableType):
        return expression.Expression(self, "or", other)

    def __ror__(self, other: OperableType):
        return expression.Expression(other, "or", self)

    def __xor__(self, other: OperableType):
        return expression.Expression(self, "xor", other)

    def __rxor__(self, other: OperableType):
        return expression.Expression(other, "xor", self)

    # <, <=, >, >=, ==, !=
    def __lt__(self, other: OperableType):
        return expression.Expression(self, "<", other)

    def __le__(self, other: OperableType):
        return expression.Expression(self, "=l=", other)

    def __gt__(self, other: OperableType):
        return expression.Expression(self, ">", other)

    def __ge__(self, other: OperableType):
        return expression.Expression(self, "=g=", other)

    # ~ -> not
    def __invert__(self):
        from gamspy._symbols.implicits import ImplicitSet

        if isinstance(self, (ImplicitSet, expression.SetExpression)):
            return expression.SetExpression(None, "not", self)
        return expression.Expression(None, "not", self)

    # a @ b
    def __matmul__(self, other):
        import gamspy._algebra.operation as operation
        from gamspy.math.matrix import _validate_matrix_mult_dims

        left_domain, right_domain, sum_domain = _validate_matrix_mult_dims(self, other)
        return operation.Sum([sum_domain], self[left_domain] * other[right_domain])

    def gamsRepr(self):
        """Representation of the symbol in GAMS"""



from __future__ import annotations

import importlib
import os
import shutil
import subprocess
import sys
from typing import TYPE_CHECKING, Annotated

import certifi
import requests
import typer

import gamspy.utils as utils
from gamspy.exceptions import GamspyException, ValidationError

from .util import add_solver_entry

if TYPE_CHECKING:
    from collections.abc import Iterable

app = typer.Typer(
    rich_markup_mode="rich",
    short_help="To install licenses and solvers.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy install license <access_code> or <path/to/license/file> | gamspy install solver <solver_name>",
    context_settings={"help_option_names": ["-h", "--help"]},
)


@app.command(
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy install license <access_code> or <path/to/license/file>",
    short_help="To install a new license",
)
def license(
    license: Annotated[
        str, typer.Argument(help="access code or path to the license file.")
    ],
    checkout_duration: int | None = typer.Option(
        None,
        "--checkout-duration",
        "-c",
        help="Specify a duration in hours to checkout a session.",
    ),
    server: str | None = typer.Option(
        "https://license.gams.com",
        "--server",
        "-s",
        help="License server adress.",
    ),
    port: int | None = typer.Option(
        None, "--port", "-p", help="Port for the license server connection."
    ),
    output: str | None = typer.Option(
        None,
        "--output",
        "-o",
        help="Specify a file path to write the license file.",
    ),
    uses_port: Annotated[
        int | None,
        typer.Option("--uses-port", help="Interprocess communication starting port."),
    ] = None,
):
    import json

    os.makedirs(utils.DEFAULT_DIR, exist_ok=True)

    is_alp = not os.path.isfile(license)

    if is_alp and len(license) != 36:
        raise ValidationError(
            f"Access code is a 36 character string or an absolute path to the "
            f"license file but {len(license)} character string ({license}) provided."
        )

    gamspy_base_dir = utils._get_gamspy_base_directory()
    license_path = os.path.join(utils.DEFAULT_DIR, "gamspy_license.txt")
    certificate_path = os.path.join(utils.DEFAULT_DIR, "gamspy_cert.crt")

    if is_alp:
        alp_id = license

        # Make cmex_type check only for GAMS license server.
        if server == "https://license.gams.com":
            params = {"access_token": alp_id}
            request = requests.get(f"{server}/license-type", params=params)
            if request.status_code != 200:
                raise ValidationError(
                    f"License server did not respond in an expected way. Request status: {request.status_code}. Reason: {request.content.decode('utf-8', errors='replace')}"
                )

            data = request.content.decode("utf-8", errors="replace")
            cmex_type = json.loads(data)["cmex_type"]
            if cmex_type not in (
                "demo",
                "community",
                "gamspy",
                "gamspy++",
                "gamsall",
            ):
                raise ValidationError(
                    f"Given access code `{alp_id} ({cmex_type})` is not valid for GAMSPy. "
                    "Make sure that you use a GAMSPy license, not a GAMS license."
                )

        command = [
            os.path.join(gamspy_base_dir, "gamsgetkey"),
            alp_id,
            "-s",
            server,
        ]

        # On-prem license server licenses start with abcdef12
        if alp_id.startswith("abcdef12"):
            command.append("-g")
            command.append(certificate_path)

        if port:
            command.append("-p")
            command.append(str(port))

        if uses_port:
            command.append("-u")
            command.append(str(uses_port))

        if checkout_duration:
            command.append("-c")
            command.append(str(checkout_duration))

        if output:
            command.append("-o")
            command.append(output)
        else:
            command.append("-o")
            command.append(license_path)

        environment_variables = os.environ.copy()
        if "CURL_CA_BUNDLE" not in environment_variables:
            environment_variables["CURL_CA_BUNDLE"] = certifi.where()

        process = subprocess.run(
            command,
            text=True,
            capture_output=True,
            encoding="utf-8",
            env=environment_variables,
        )
        if process.returncode:
            raise ValidationError(process.stderr)

        license_file_path = output if output else license_path
        with open(license_file_path) as file:
            license_text = file.read()

        lines = license_text.splitlines()
        try:
            license_type = lines[0][54]
        except (Exception, IndexError) as e:
            raise ValidationError(
                f"Invalid license text. \nLicense text: {license_text}\nstdout: {process.stdout}\nstderr: {process.stderr}"
            ) from e

        if license_type == "+":
            if lines[2][:2] not in ("00", "07", "08", "09"):
                raise ValidationError(
                    f"Given access code `{alp_id}` is not valid for GAMSPy. "
                    "Make sure that you use a GAMSPy license, not a GAMS license."
                )
        else:
            if lines[2][8:10] not in ("00", "07", "08", "09"):
                raise ValidationError(
                    f"Given access code `{alp_id}` is not valid for GAMSPy. "
                    "Make sure that you use a GAMSPy license, not a GAMS license."
                )
    else:
        with open(license) as file:
            lines = file.read().splitlines()

        license_type = lines[0][54]
        if license_type == "+":
            if lines[2][:2] not in ("00", "07", "08", "09"):
                raise ValidationError(
                    f"Given license file `{license}` is not valid for GAMSPy. "
                    "Make sure that you use a GAMSPy license, not a GAMS license."
                )
        else:
            if lines[2][8:10] not in ("00", "07", "08", "09"):
                raise ValidationError(
                    f"Given license file `{license}` is not valid for GAMSPy. "
                    "Make sure that you use a GAMSPy license, not a GAMS license."
                )

        shutil.copy(license, license_path)


def append_dist_info(files, gamspy_base_dir: str):
    """Updates dist-info/RECORD in site-packages for pip uninstall"""
    import gamspy as gp

    gamspy_path: str = gp.__path__[0]
    dist_info_path = f"{gamspy_path}-{gp.__version__}.dist-info"

    with open(dist_info_path + os.sep + "RECORD", "a", encoding="utf-8") as record:
        gamspy_base_relative_path = os.sep.join(gamspy_base_dir.split(os.sep)[-3:])

        lines = []
        for file in files:
            line = f"{gamspy_base_relative_path}{os.sep}{file},,"
            lines.append(line)

        record.write("\n".join(lines) + "\n")


@app.command(
    short_help="To install solvers",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy install solver <solver_name>",
)
def solver(
    solver: list[str] = typer.Argument(  # noqa: B008
        None,
        help="solver names to be installed",
        autocompletion=lambda: [s.lower() for s in utils.getAvailableSolvers()],
    ),
    install_all_solvers: bool = typer.Option(
        False,
        "--install-all-solvers",
        "--all",
        "-a",
        help="Installs all available add-on solvers.",
    ),
    existing_solvers: bool = typer.Option(
        False,
        "--existing-solvers",
        help="Reinstalls previously installed add-on solvers.",
    ),
    skip_pip_install: bool = typer.Option(
        False,
        "--skip-pip-install",
        "-s",
        help="If you already have the solver installed, skip pip install and update gamspy installed solver list.",
    ),
    use_uv: bool = typer.Option(
        False, "--use-uv", help="Use uv instead of pip to install solvers."
    ),
):
    try:
        import gamspy_base
    except ModuleNotFoundError as e:
        e.msg = "You must first install gamspy_base to use this functionality"
        raise e

    addons_path = os.path.join(utils.DEFAULT_DIR, "solvers.txt")
    os.makedirs(utils.DEFAULT_DIR, exist_ok=True)
    environment_variables = os.environ.copy()
    if "CURL_CA_BUNDLE" not in environment_variables:
        environment_variables["CURL_CA_BUNDLE"] = certifi.where()

    def install_addons(addons: Iterable[str]):
        for item in addons:
            solver_name = item.lower()

            if solver_name.upper() in gamspy_base.default_solvers:
                print(f"`{solver_name}` is a default solver, skipping...")
                continue
            installable_solvers = utils.getInstallableSolvers(gamspy_base.directory)
            if solver_name.upper() not in installable_solvers:
                raise ValidationError(
                    f'Given solver name ("{solver_name}") is not valid. Available'
                    f" solvers that can be installed: {installable_solvers}"
                )

            installed_solvers = utils.getInstalledSolvers(gamspy_base.directory)
            if solver_name.upper() in installed_solvers:
                print(f"`{solver_name}` is already installed, skipping...")
                continue

            if not skip_pip_install:
                solver_version = gamspy_base.__version__
                # install specified solver
                if use_uv:
                    command = [
                        "uv",
                        "pip",
                        "install",
                        f"gamspy-{solver_name}=={solver_version}",
                        "--force-reinstall",
                    ]
                else:
                    command = [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        f"gamspy-{solver_name}=={solver_version}",
                        "--force-reinstall",
                    ]
                try:
                    _ = subprocess.run(
                        command,
                        check=True,
                        encoding="utf-8",
                        stderr=subprocess.PIPE,
                        env=environment_variables,
                    )
                except subprocess.CalledProcessError as e:
                    raise GamspyException(
                        f"Could not install gamspy-{solver_name}. Please check your internet connection. If it's not related to your internet connection, PyPI servers might be down. Please retry it later. Here is the error message of pip:\n\n{e.stderr}"
                    ) from e
            else:
                try:
                    solver_lib = importlib.import_module(f"gamspy_{solver_name}")
                except ModuleNotFoundError as e:
                    e.msg = f"You must install gamspy-{solver_name} first!"
                    raise e

            # copy solver files to gamspy_base
            gamspy_base_dir = utils._get_gamspy_base_directory()
            solver_lib = importlib.import_module(f"gamspy_{solver_name}")

            file_paths = solver_lib.file_paths
            for file in file_paths:
                shutil.copy(file, gamspy_base_dir)

            files = solver_lib.files
            verbatims = [solver_lib.verbatim]
            append_dist_info(files, gamspy_base_dir)
            add_solver_entry(gamspy_base_dir, solver_name, verbatims)

            try:
                with open(addons_path) as file:
                    installed = file.read().splitlines()
                    installed = [
                        solver for solver in installed if solver not in {"", "\n"}
                    ]
            except FileNotFoundError:
                installed = []

            with open(addons_path, "w") as file:
                if solver_name.upper() not in installed:
                    file.write("\n".join(installed + [solver_name.upper()]))

    if install_all_solvers:
        available_solvers = utils.getAvailableSolvers()
        installed_solvers = utils.getInstalledSolvers(gamspy_base.directory)
        diff = []
        for available_solver in available_solvers:
            if available_solver not in installed_solvers:
                diff.append(available_solver)

        install_addons(diff)
        return

    if existing_solvers:
        try:
            with open(addons_path) as file:
                solvers = file.read().splitlines()
                solvers = [solver for solver in solvers if solver not in {"", "\n"}]
                install_addons(solvers)
                return
        except FileNotFoundError as e:
            raise ValidationError("No existing add-on solvers found!") from e

    if solver is None:
        raise ValidationError(
            "Solver name is missing: `gamspy install solver <solver_name>`"
        )

    install_addons(solver)


if __name__ == "__main__":
    app()



from __future__ import annotations

from typing import TYPE_CHECKING

from gamspy.formulations.nn.mpool2d import _MPool2d

if TYPE_CHECKING:
    import gamspy as gp


class MaxPool2d(_MPool2d):
    """
    Formulation generator for 2D Max Pooling in GAMS.

    Parameters
    ----------
    container : Container
        Container that will contain the new variable and equations.
    kernel_size : int | tuple[int, int]
        Filter size
    stride : int | tuple[int, int] | None
        Stride in the max pooling, it is equal to kernel_size if not provided
    padding : int | tuple[int, int]
        Amount of padding to be added to input, by default 0
    name_prefix : str | None
        Prefix for generated GAMSPy symbols, by default None which means
        random prefix. Using the same name_prefix in different formulations causes name
        conflicts. Do not use the same name_prefix again.

    Examples
    --------
    >>> import gamspy as gp
    >>> from gamspy.math import dim
    >>> m = gp.Container()
    >>> # 2x2 max pooling
    >>> mp1 = gp.formulations.MaxPool2d(m, (2, 2))
    >>> inp = gp.Variable(m, domain=dim((10, 1, 24, 24)))
    >>> out, eqs = mp1(inp)
    >>> type(out)
    <class 'gamspy._symbols.variable.Variable'>
    >>> [len(x) for x in out.domain]
    [10, 1, 12, 12]

    """

    def __init__(
        self,
        container: gp.Container,
        kernel_size: int | tuple[int, int],
        stride: int | None = None,
        padding: int = 0,
        name_prefix: str | None = None,
    ):
        super().__init__("max", container, kernel_size, stride, padding, name_prefix)

    def __call__(
        self,
        input: gp.Parameter | gp.Variable,
        big_m: int = 1000,
        propagate_bounds: bool = True,
    ) -> tuple[gp.Variable, list[gp.Equation]]:
        """
        Forward pass your input, generate output and equations required for
        calculating the max pooling. Returns the output variable and the list
        of equations required for the max pooling formulation. if propagate_bounds
        is True, it will also set the bounds for the output variable based on the input.
        It will also compute the big M value required for the pooling operation
        using the bounds.

        Parameters
        ----------
        input : gp.Parameter | gp.Variable
                input to the max pooling 2d layer, must be in shape
                (batch x in_channels x height x width)
        big_m: int
               Big M value that is required for the pooling operation.
               Default value: 1000.
        propagate_bounds: bool
                If True, it will set the bounds for the output variable
                based on the input.
                Default value: True

        Returns
        -------
        tuple[gp.Variable, list[gp.Equation]]

        """
        return super().__call__(input, big_m, propagate_bounds)



from gamspy._symbols.alias import Alias
from gamspy._symbols.equation import Equation, EquationType
from gamspy._symbols.parameter import Parameter
from gamspy._symbols.set import Set
from gamspy._symbols.universe_alias import UniverseAlias
from gamspy._symbols.variable import Variable, VariableType

__all__ = [
    "Alias",
    "Equation",
    "EquationType",
    "Parameter",
    "Set",
    "UniverseAlias",
    "Variable",
    "VariableType",
]



from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np


@dataclass(frozen=True)
class DecisionTreeStruct:
    """
    Represents the components of `sklearn.tree`.

    This dataclass stores the core arrays (like children, features, thresholds,
    and values) that define the tree's architecture and decision rules.

    Attributes
    ----------
    children_left: np.ndarray
        An array where `children_left[i]` is the ID
        of the left child of node `i`. Leaf nodes have -1.
        Defaults to an empty numpy array.
    children_right: np.ndarray
        An array where `children_right[i]` is the ID
        of the right child of node `i`. Leaf nodes have -1.
        Defaults to an empty numpy array.
    feature: np.ndarray
        An array where `feature[i]` is the index of the
        feature used for splitting at node `i`. Leaf nodes have -2.
        Defaults to an empty numpy array.
    threshold: np.ndarray
        An array where `threshold[i]` is the threshold
        value used for splitting at node `i` based on `feature[i]`.
        Leaf nodes have -2.0. Defaults to an empty numpy array.
    value: np.ndarray
        An array (typically 2D for scikit-learn trees,
        squeezed to 1D for single-output regressors) where `value[i]`
        contains the prediction value(s) for node `i`. For leaf nodes,
        this is the final prediction. Defaults to an empty numpy array.
    capacity : int
        The total number of nodes allocated in the underlying
        tree structure arrays. Defaults to 0.
    n_features : int
        The number of features the decision tree was trained on
        or expects as input. Defaults to 0.
    """

    children_left: np.ndarray = field(default_factory=lambda: np.array([]), repr=False)
    children_right: np.ndarray = field(default_factory=lambda: np.array([]), repr=False)
    feature: np.ndarray = field(default_factory=lambda: np.array([]), repr=False)
    threshold: np.ndarray = field(default_factory=lambda: np.array([]), repr=False)
    value: np.ndarray = field(default_factory=lambda: np.array([]), repr=False)
    capacity: int = 0
    n_features: int = 0

    def __repr__(self):
        def arr_info(arr):
            return f"shape={arr.shape}, dtype={arr.dtype}" if arr.size else "empty"

        return (
            f"{self.__class__.__name__}("
            f"children_left={arr_info(self.children_left)}, "
            f"children_right={arr_info(self.children_right)}, "
            f"feature={arr_info(self.feature)}, "
            f"threshold={arr_info(self.threshold)}, "
            f"value={arr_info(self.value)}, "
            f"capacity={self.capacity}, "
            f"n_features={self.n_features})"
        )

    def __str__(self):
        return (
            f"{self.__class__.__name__}: "
            f"{self.capacity} nodes, {self.n_features} features, "
            f"arrays: "
            f"children_left({self.children_left.size}), "
            f"children_right({self.children_right.size}), "
            f"feature({self.feature.size}), "
            f"threshold({self.threshold.size}), "
            f"value({self.value.size})"
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import gamspy._algebra.condition as condition
import gamspy._algebra.domain as domain
import gamspy._algebra.expression as expression
import gamspy._algebra.operable as operable
import gamspy._symbols as syms
import gamspy._symbols.implicits as implicits
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd

    from gamspy._algebra import Domain
    from gamspy._algebra.condition import Condition
    from gamspy._algebra.expression import Expression
    from gamspy._symbols import Alias, Parameter, Set
    from gamspy._symbols.implicits import (
        ImplicitParameter,
        ImplicitSet,
        ImplicitVariable,
    )
    from gamspy.math.misc import MathOp


class Operation(operable.Operable):
    def __init__(
        self,
        domain: Set | Alias | ImplicitSet | Sequence[Set | Alias] | Domain | Condition,
        rhs: (
            Expression
            | Operation
            | MathOp
            | ImplicitSet
            | ImplicitVariable
            | ImplicitParameter
            | int
            | bool
        ),
        op_name: str,
    ):
        self.op_domain = utils._to_list(domain)  # type: ignore
        assert len(self.op_domain) > 0, "Operation requires at least one index"
        self.rhs = rhs
        self._op_name = op_name
        self.raw_domain = self._get_raw_domain()

        # allow conditions
        self.where = condition.Condition(self)

        self._bare_op_domain = utils._get_set(self.op_domain)
        self.container = self.raw_domain[0].container
        self.domain: list[Set | Alias] = []

        self._operation_indices = []
        if isinstance(rhs, condition.Condition):
            rhs = rhs.conditioning_on  # type: ignore

        if not isinstance(rhs, (bool, float, int)):
            for i, x in enumerate(rhs.domain):  # type: ignore
                try:
                    sum_index = self._bare_op_domain.index(x)
                    self._operation_indices.append((i, sum_index))
                except ValueError:
                    self.domain.append(x)

        self.dimension: int = validation.get_dimension(self.domain)
        controlled_domain = list(self._bare_op_domain)
        controlled_domain.extend(getattr(rhs, "controlled_domain", []))
        self.controlled_domain = list(set(controlled_domain))

    def __getitem__(self, indices: Sequence | str):
        domain = validation.validate_domain(self, indices)
        for index, sum_index in self._operation_indices:
            domain.insert(index, self._bare_op_domain[sum_index])

        if isinstance(self.rhs, (bool, float, int)):
            return Operation(self.op_domain, self.rhs, self._op_name)

        return Operation(self.op_domain, self.rhs[domain], self._op_name)  # type: ignore

    @property
    def records(self) -> pd.DataFrame | None:
        """
        Evaluates the operation and returns the resulting records.

        Returns
        -------
        pd.DataFrame | None
        """
        assert self.container is not None
        temp_name = "a" + utils._get_unique_name()
        domain: list[Set | Alias] = []
        for elem in self.domain:
            if hasattr(elem, "dimension") and elem.dimension > 1:
                domain.extend(elem.domain)
            else:
                domain.append(elem)

        temp_param = syms.Parameter._constructor_bypass(
            self.container, temp_name, domain
        )
        temp_param[...] = self
        del self.container.data[temp_name]
        return temp_param.records

    def toValue(self) -> float | None:
        """
        Convenience method to return the records of the operation as a Python float. Only possible if there is a single record as a result of the operation.

        Returns
        -------
        float | None
        """
        records = self.records
        if records is not None:
            return records["value"][0]

        return records

    def toList(self) -> list | None:
        """
        Convenience method to return the records of the operation as a list.

        Returns
        -------
        list | None
        """
        records = self.records
        if records is not None:
            return records.values.tolist()

        return None

    def _get_raw_domain(self) -> list[Set | Alias | ImplicitSet]:
        raw_domain = []
        for elem in self.op_domain:
            if isinstance(elem, condition.Condition):
                if isinstance(elem.conditioning_on, implicits.ImplicitSet):
                    raw_domain.append(elem.conditioning_on.parent)
                elif isinstance(elem.conditioning_on, (syms.Set, syms.Alias)):
                    raw_domain.append(elem.conditioning_on)
                elif isinstance(elem.conditioning_on, domain.Domain):
                    raw_domain += elem.conditioning_on.sets
            elif isinstance(elem, domain.Domain):
                raw_domain += elem.sets
            elif isinstance(elem, implicits.ImplicitSet):
                raw_domain.append(elem)
            else:
                raw_domain.append(elem)

        return raw_domain

    def _validate_operation(
        self, control_stack: list[Set | Alias | ImplicitSet]
    ) -> None:
        for elem in self.raw_domain:
            if isinstance(elem, implicits.ImplicitSet):
                control_stack += [
                    member for member in elem.domain if member not in control_stack
                ] + [elem.parent]

            if elem in control_stack:
                raise ValidationError(f"Set {elem} is already in control")

        stack = control_stack + self.raw_domain
        if isinstance(self.rhs, expression.Expression):
            self.rhs._validate_definition(utils._unpack(stack))
        elif isinstance(self.rhs, Operation):
            self.rhs._validate_operation(utils._unpack(stack))

    def _get_index_str(self) -> str:
        if len(self.op_domain) == 1:
            return self.op_domain[0].gamsRepr()

        return "(" + ",".join([index.gamsRepr() for index in self.op_domain]) + ")"

    def __eq__(self, other):
        return expression.Expression(self, "=e=", other)

    def __ne__(self, other):
        return expression.Expression(self, "ne", other)

    def _replace_operations(self, output: str) -> str:
        output = output.replace("=l=", "<=")
        output = output.replace("=g=", ">=")
        output = output.replace("=e=", "eq")

        return output

    def gamsRepr(self) -> str:
        # Ex: sum((i,j), c(i,j) * x(i,j))
        output = f"{self._op_name}("

        index_str = self._get_index_str()

        output += index_str
        output += ","

        if isinstance(self.rhs, float):
            self.rhs = utils._map_special_values(self.rhs)

        if isinstance(self.rhs, bool):
            self.rhs = (
                "yes" if self.rhs is True else "no"  # type: ignore
            )

        expression_str = (
            str(self.rhs)
            if isinstance(self.rhs, (bool, float, int, str))
            else self.rhs.gamsRepr()
        )

        output += expression_str
        output += ")"

        output = self._replace_operations(output)

        return output

    def latexRepr(self) -> str:
        """
        Representation of this operation in Latex.

        Returns
        -------
        str
        """
        op_map = {
            "sum": "sum",
            "prod": "prod",
            "smax": "max",
            "smin": "min",
            "sand": "sand",
            "sor": "sor",
        }

        indices = []
        given_condition = None
        for index in self.op_domain:
            if isinstance(index, condition.Condition):
                indices.append(index.conditioning_on)
                given_condition = index.condition
            else:
                indices.append(index)

        index_str = ",".join([index.latexRepr() for index in indices])

        if given_condition is not None:
            condition_str = str(given_condition)
            if hasattr(given_condition, "latexRepr"):
                condition_str = given_condition.latexRepr()
            index_str += " ~ | ~ " + condition_str

        expression_str = (
            str(self.rhs)
            if isinstance(self.rhs, (int, float, str))
            else self.rhs.latexRepr()
        )
        representation = f"\\{op_map[self._op_name]}_{{{index_str}}} {expression_str}"
        return representation


class Sum(Operation):
    """
    Represents a sum operation over a domain.

    Parameters
    ----------
    domain : Set | Alias | ImplicitSet | Sequence[Set | Alias], Domain, Condition
    expression : (
            Expression
            | MathOp
            | ImplicitVariable
            | ImplicitParameter
            | int
            | bool
            | Variable
            | Parameter
            | Operation
        )

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, "i", records=['i1','i2', 'i3'])
    >>> v = gp.Variable(m, "v")
    >>> e = gp.Equation(m, "e", type="eq")
    >>> d = gp.Parameter(m, "d", domain=[i], records=[("i1", 1), ("i2", 2), ("i3", 4)])
    >>> e[...] = gp.Sum(i, d[i]) <= v

    """

    def __init__(
        self,
        domain: Set | Alias | ImplicitSet | Sequence[Set | Alias] | Domain | Condition,
        expression: Operation
        | Expression
        | MathOp
        | ImplicitSet
        | ImplicitParameter
        | ImplicitVariable
        | int
        | bool,
    ):
        super().__init__(domain, expression, "sum")

    def __repr__(self) -> str:
        return f"Sum(domain={self.domain}, expression={self.rhs})"

    def gamsRepr(self):
        """
        Representation of the Sum operation in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> from gamspy import Container, Set, Parameter, Variable, Sum
        >>> m = Container()
        >>> i = Set(m, "i", records=['i1','i2', 'i3'])
        >>> c = Parameter(m, "c", domain=i)
        >>> v = Variable(m, "v", domain=i)
        >>> Sum(i, c[i]*v[i]).gamsRepr()
        'sum(i,c(i) * v(i))'

        """
        repr = super().gamsRepr()
        return repr


class Product(Operation):
    """
    Represents a product operation over a domain.

    Parameters
    ----------
    domain : Set | Alias | Sequence[Set | Alias], Domain, Expression
    expression : (
            Expression
            | MathOp
            | ImplicitVariable
            | ImplicitParameter
            | int
            | bool
            | Variable
            | Parameter
            | Operation
        )

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, "i", records=['i1','i2', 'i3'])
    >>> v = gp.Variable(m, "v")
    >>> e = gp.Equation(m, "e", type="eq")
    >>> p = gp.Parameter(m, "p", domain=[i], records=[("i1", 1), ("i2", 2), ("i3", 4)])
    >>> e[...] = gp.Product(i, p[i]) <= v

    """

    def __init__(
        self,
        domain: Set | Alias | ImplicitSet | Sequence[Set | Alias] | Domain | Condition,
        expression: Operation
        | Expression
        | MathOp
        | ImplicitSet
        | ImplicitParameter
        | ImplicitVariable
        | int
        | bool,
    ):
        super().__init__(domain, expression, "prod")

    def __repr__(self) -> str:
        return f"Product(domain={self.domain}, expression={self.rhs})"

    def gamsRepr(self):
        """
        Representation of the Product operation in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> from gamspy import Container, Set, Parameter, Variable, Product
        >>> m = Container()
        >>> i = Set(m, "i", records=['i1','i2', 'i3'])
        >>> c = Parameter(m, "c", domain=i)
        >>> v = Variable(m, "v", domain=i)
        >>> Product(i, c[i]*v[i]).gamsRepr()
        'prod(i,c(i) * v(i))'

        """
        repr = super().gamsRepr()
        return repr


class Smin(Operation):
    """
    Represents a smin operation over a domain.

    Parameters
    ----------
    domain : Set | Alias | Sequence[Set | Alias], Domain, Expression
    expression : (
            Expression
            | MathOp
            | ImplicitVariable
            | ImplicitParameter
            | int
            | bool
            | Variable
            | Parameter
            | Operation
        )

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, "i", records=['i1','i2', 'i3'])
    >>> v = gp.Variable(m, "v")
    >>> e = gp.Equation(m, "e", type="eq")
    >>> p = gp.Parameter(m, "p", domain=[i], records=[("i1", 1), ("i2", 2), ("i3", 4)])
    >>> e[...] = gp.Smin(i, p[i]) <= v

    """

    def __init__(
        self,
        domain: Set | Alias | ImplicitSet | Sequence[Set | Alias] | Domain | Condition,
        expression: Operation
        | Expression
        | MathOp
        | ImplicitSet
        | ImplicitParameter
        | ImplicitVariable
        | int
        | bool,
    ):
        super().__init__(domain, expression, "smin")

    def __repr__(self) -> str:
        return f"Smin(domain={self.domain}, expression={self.rhs})"

    def gamsRepr(self):
        """
        Representation of the Smin operation in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> from gamspy import Container, Set, Parameter, Variable, Smin
        >>> m = Container()
        >>> i = Set(m, "i", records=['i1','i2', 'i3'])
        >>> c = Parameter(m, "c", domain=i)
        >>> v = Variable(m, "v", domain=i)
        >>> Smin(i, c[i]*v[i]).gamsRepr()
        'smin(i,c(i) * v(i))'

        """
        repr = super().gamsRepr()
        return repr


class Smax(Operation):
    """
    Represents a smax operation over a domain.

    Parameters
    ----------
    domain : Set | Alias | Sequence[Set | Alias], Domain, Expression
    expression : (
            Expression
            | MathOp
            | ImplicitVariable
            | ImplicitParameter
            | int
            | bool
            | Variable
            | Parameter
            | Operation
        )

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, "i", records=['i1','i2', 'i3'])
    >>> v = gp.Variable(m, "v")
    >>> e = gp.Equation(m, "e", type="eq")
    >>> p = gp.Parameter(m, "p", domain=[i], records=[("i1", 1), ("i2", 2), ("i3", 4)])
    >>> e[...] = gp.Smax(i, p[i]) <= v

    """

    def __init__(
        self,
        domain: Set | Alias | ImplicitSet | Sequence[Set | Alias] | Domain | Condition,
        expression: Operation
        | Expression
        | MathOp
        | ImplicitSet
        | ImplicitParameter
        | ImplicitVariable
        | int
        | bool,
    ):
        super().__init__(domain, expression, "smax")

    def __repr__(self) -> str:
        return f"Smax(domain={self.domain}, expression={self.rhs})"

    def gamsRepr(self):
        """
        Representation of the Smax operation in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> from gamspy import Container, Set, Parameter, Variable, Smax
        >>> m = Container()
        >>> i = Set(m, "i", records=['i1','i2', 'i3'])
        >>> c = Parameter(m, "c", domain=i)
        >>> v = Variable(m, "v", domain=i)
        >>> Smax(i, c[i]*v[i]).gamsRepr()
        'smax(i,c(i) * v(i))'

        """
        repr = super().gamsRepr()
        return repr


class Sand(Operation):
    """
    Represents a sand operation over a domain.

    Parameters
    ----------
    domain : Set | Alias | Sequence[Set | Alias], Domain, Expression
    expression : (
            Expression
            | MathOp
            | ImplicitVariable
            | ImplicitParameter
            | int
            | bool
            | Variable
            | Parameter
            | Operation
        )

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, "i", records=['i1','i2', 'i3'])
    >>> p = gp.Parameter(m, domain=i)
    >>> p[i] = gp.math.uniformInt(0,1)
    >>> result = gp.Parameter(m)
    >>> result[:] = gp.Sand(i, p[i])

    """

    def __init__(
        self,
        domain: Set | Alias | ImplicitSet | Sequence[Set | Alias] | Domain | Condition,
        expression: Operation
        | Expression
        | MathOp
        | ImplicitSet
        | ImplicitParameter
        | ImplicitVariable
        | int
        | bool,
    ):
        super().__init__(domain, expression, "sand")

    def __repr__(self) -> str:
        return f"Sand(domain={self.domain}, expression={self.rhs})"

    def gamsRepr(self):
        """
        Representation of the Sand operation in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> from gamspy import Container, Set, Parameter, Variable, Sor, Sand
        >>> m = Container()
        >>> i = Set(m, "i", records=['i1','i2', 'i3'])
        >>> v = Variable(m, "v", domain=i, type="binary")
        >>> Sand(i, v[i]).gamsRepr()
        'sand(i,v(i))'

        """
        repr = super().gamsRepr()
        return repr


class Sor(Operation):
    """
    Represents a sor operation over a domain.

    Parameters
    ----------
    domain : Set | Alias | Sequence[Set | Alias], Domain, Expression
    expression : (
            Expression
            | MathOp
            | ImplicitVariable
            | ImplicitParameter
            | int
            | bool
            | Variable
            | Parameter
            | Operation
        )

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, "i", records=['i1','i2', 'i3'])
    >>> p = gp.Parameter(m, domain=i)
    >>> p[i] = gp.math.uniformInt(0,1)
    >>> result = gp.Parameter(m)
    >>> result[:] = gp.Sor(i, p[i])

    """

    def __init__(
        self,
        domain: Set | Alias | ImplicitSet | Sequence[Set | Alias] | Domain | Condition,
        expression: Operation
        | Expression
        | MathOp
        | ImplicitSet
        | ImplicitParameter
        | ImplicitVariable
        | int
        | bool,
    ):
        super().__init__(domain, expression, "sor")

    def __repr__(self) -> str:
        return f"Sor(domain={self.domain}, expression={self.rhs})"

    def gamsRepr(self):
        """
        Representation of the Sor operation in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> from gamspy import Container, Set, Parameter, Variable, Sor
        >>> m = Container()
        >>> i = Set(m, "i", records=['i1','i2', 'i3'])
        >>> v = Variable(m, "v", domain=i, type="binary")
        >>> Sor(i, v[i]).gamsRepr()
        'sor(i,v(i))'

        """
        repr = super().gamsRepr()
        return repr


class Ord(operable.Operable):
    """
    The operator ord may be used only with one-dimensional sets.

    Parameters
    ----------
    set : Set | Alias

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> t = gp.Set(m, name="t", description="time periods", records=[str(x) for x in range(1985, 1996)])
    >>> val = gp.Parameter(m, name="val", domain=[t])
    >>> val[t] = gp.Ord(t)

    """

    def __new__(cls, symbol: Set | Alias):
        if not isinstance(symbol, (syms.Set, syms.Alias)):
            raise ValidationError("Ord operation is only for Set and Alias objects!")

        if symbol.is_singleton:
            return 1

        return super().__new__(cls)

    def __init__(self, symbol: Set | Alias):
        self._symbol = symbol
        self.container = symbol.container
        self.domain: list[Set | Alias] = []
        self.where = condition.Condition(self)

    def __eq__(self, other):
        return expression.Expression(self, "eq", other)

    def __ge__(self, other):
        return expression.Expression(self, ">=", other)

    def __le__(self, other):
        return expression.Expression(self, "<=", other)

    def __ne__(self, other):
        return expression.Expression(self, "ne", other)

    def gamsRepr(self) -> str:
        """
        Representation of the Ord operation in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> from gamspy import Container, Set, Ord
        >>> m = Container()
        >>> i = Set(m, "i", records=['i1','i2', 'i3'])
        >>> Ord(i).gamsRepr()
        'ord(i)'

        """
        return f"ord({self._symbol.name})"

    def latexRepr(self) -> str:
        """
        Representation of Ord function in Latex.

        Returns
        -------
        str
        """
        return f"ord({self._symbol._latex_name})"


class Card(operable.Operable):
    """
    The operator card may be used with any symbol and returns its number of records.

    Parameters
    ----------
    symbol : Set | Alias | Parameter | Variable | Equation | Model

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> t = gp.Set(m, name="t", description="time periods", records=[str(x) for x in range(1985, 1996)])
    >>> s = gp.Parameter(m, name="s")
    >>> s[...] = gp.Card(t)

    """

    def __init__(
        self,
        symbol: Set | Alias | Parameter,
    ) -> None:
        if not isinstance(symbol, (syms.Set, syms.Alias, syms.Parameter)):
            raise ValidationError(
                "Card operation is only for Set, Alias and Parameter objects!"
            )

        self._symbol = symbol
        self.container = symbol.container
        self.domain: list[Set | Alias] = []
        self.where = condition.Condition(self)

    def __eq__(self, other) -> Expression:  # type: ignore
        return expression.Expression(self, "eq", other)

    def __ge__(self, other):
        return expression.Expression(self, ">=", other)

    def __le__(self, other):
        return expression.Expression(self, "<=", other)

    def __ne__(self, other):  # type: ignore
        return expression.Expression(self, "ne", other)

    def __bool__(self):
        raise ValidationError(
            "Card operation cannot be used as a truth value. Use len(<symbol>.records) instead."
        )

    def gamsRepr(self) -> str:
        """
        Representation of the Card operation in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> from gamspy import Container, Set, Card
        >>> m = Container()
        >>> i = Set(m, "i", records=['i1','i2', 'i3'])
        >>> Card(i).gamsRepr()
        'card(i)'

        """
        return f"card({self._symbol.name})"

    def latexRepr(self) -> str:
        """
        Representation of Card function in Latex.

        Returns
        -------
        str
        """
        return f"card({self._symbol._latex_name})"



from __future__ import annotations

import os
import threading
from typing import TYPE_CHECKING, Any

import gams.transfer as gt
from gams.core.gdx import GMS_DT_ALIAS

import gamspy as gp
import gamspy._algebra.condition as condition
import gamspy._algebra.expression as expression
import gamspy._algebra.operable as operable
import gamspy._symbols.implicits as implicits
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy._symbols.set import SetMixin
from gamspy._symbols.symbol import Symbol
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from collections.abc import Sequence

    from gamspy import Container, Set
    from gamspy._algebra.expression import Expression


class Alias(gt.Alias, operable.Operable, Symbol, SetMixin):
    """
    Represents an Alias symbol in GAMS.
    https://gamspy.readthedocs.io/en/latest/user/basics/alias.html

    Parameters
    ----------
    container : Container
        Container of the alias.
    name : str, optional
        Name of the alias.
    alias_with : Set
        Alias set object.

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, "i")
    >>> j = gp.Alias(m, "j", i)

    """

    @classmethod
    def _constructor_bypass(
        cls, container: Container, name: str, alias_with: Set | Alias
    ):
        # create new symbol object
        obj = Alias.__new__(cls, container, name, alias_with)

        # set private properties directly
        obj._requires_state_check = False
        obj._container = container
        container._requires_state_check = True
        obj._name = name
        obj._alias_with = alias_with
        obj._modified = True

        # typing
        obj._gams_type = GMS_DT_ALIAS
        obj._gams_subtype = 1

        # add to container
        container.data.update({name: obj})

        # gamspy attributes
        obj.where = condition.Condition(obj)
        obj._latex_name = name.replace("_", r"\_")
        obj.container._add_statement(obj)
        obj._metadata = {}

        return obj

    def __new__(
        cls,
        container: Container | None = None,
        name: str | None = None,
        alias_with: Set | Alias | None = None,
    ):
        if container is not None and not isinstance(container, gp.Container):
            raise TypeError(
                f"Container must of type `Container` but found {type(container)}"
            )

        if name is None:
            return object.__new__(cls)
        else:
            if not isinstance(name, str):
                raise TypeError(f"Name must of type `str` but found {type(name)}")
            try:
                if not container:
                    container = gp._ctx_managers[
                        (os.getpid(), threading.get_native_id())
                    ]

                symbol = container[name]
                if isinstance(symbol, cls):
                    return symbol

                raise TypeError(
                    f"Cannot overwrite symbol `{name}` in container"
                    " because it is not an Alias object)"
                )
            except KeyError:
                return object.__new__(cls)

    def __init__(
        self,
        container: Container | None = None,
        name: str | None = None,
        alias_with: Set | Alias = None,  # type: ignore
    ):
        self._metadata: dict[str, Any] = {}
        self._assignment: Expression | None = None
        # does symbol exist
        has_symbol = False
        if isinstance(getattr(self, "container", None), gp.Container):
            has_symbol = True

        if has_symbol:
            # reset some properties
            self._requires_state_check = True
            self.container._requires_state_check = True
            self._modified = True
            self.alias_with = alias_with
        else:
            if container is None:
                try:
                    container = gp._ctx_managers[
                        (os.getpid(), threading.get_native_id())
                    ]
                except KeyError as e:
                    raise ValidationError("Alias requires a container.") from e
            assert container is not None

            if name is not None:
                name = validation.validate_name(name)
            else:
                name = container._get_symbol_name(prefix="a")

            super().__init__(container, name, alias_with)
            self._latex_name = self.name.replace("_", r"\_")

            validation.validate_container(self, self.domain)
            self.where = condition.Condition(self)
            self.container._add_statement(self)

            self._modified = False
            self.container._synch_with_gams()

    def _serialize(self) -> dict:
        info: dict[str, Any] = {"_metadata": self._metadata}
        if self._assignment is not None:
            info["_assignment"] = self._assignment.getDeclaration()

        return info

    def _deserialize(self, info: dict) -> None:
        for key, value in info.items():
            if key == "_assignment":
                left, right = value.split(" = ")
                value = expression.Expression(left, "=", right)

            setattr(self, key, value)

    def __len__(self):
        if self.records is not None:
            return len(self.records.index)

        return 0

    def __bool__(self):
        raise ValidationError(
            "Alias cannot be used as a truth value. Use len(<symbol>.records) instead."
        )

    def __repr__(self) -> str:
        return f"Alias(name='{self.name}', alias_with={self.alias_with})"

    def __getitem__(self, indices: Sequence | str) -> implicits.ImplicitSet:
        domain = validation.validate_domain(self, indices)

        return implicits.ImplicitSet(self, name=self.name, domain=domain)

    def __setitem__(self, indices: tuple | str, rhs):
        # self[domain] = rhs
        domain = validation.validate_domain(self, indices)

        if isinstance(rhs, bool):
            rhs = "yes" if rhs is True else "no"  # type: ignore

        statement = expression.Expression(
            implicits.ImplicitSet(self, name=self.name, domain=domain),
            "=",
            rhs,
        )

        statement._validate_definition(utils._unpack(domain))

        self.container._add_statement(statement)
        self._assignment = statement

        if self.alias_with.synchronize:  # type: ignore
            self.container._synch_with_gams(gams_to_gamspy=True)

    def _setRecords(self, records: Any, *, uels_on_axes: bool = False) -> None:
        super().setRecords(records, uels_on_axes)

        if gp.get_option("DROP_DOMAIN_VIOLATIONS"):
            if self.hasDomainViolations():
                self._domain_violations = self.getDomainViolations()
                self.dropDomainViolations()
            else:
                self._domain_violations = None

    @property
    def synchronize(self):
        raise ValidationError(
            "Each Alias object is tied to a Set. Change the synchronization setting of the set instead."
        )

    @synchronize.setter
    def synchronize(self, value: bool):
        assert self.alias_with is not None
        raise ValidationError(
            f"Alias `{self.name}` object is tied to a Set `{self.alias_with.name}`."
            f"Change the synchronization setting of the Set `{self.alias_with.name}` instead."
        )

    def gamsRepr(self) -> str:
        """
        Representation of this Alias in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", domain=["*"], records=['i1','i2'])
        >>> j = gp.Alias(m, "j", i)
        >>> j.gamsRepr()
        'j'

        """
        return self.name

    def getDeclaration(self):
        """
        Declaration of the Alias in GAMS

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1','i2'])
        >>> j = gp.Alias(m, "j", i)
        >>> j.getDeclaration()
        'Alias(i,j);'

        """
        return f"Alias({self.alias_with.name},{self.name});"

    def getAssignment(self) -> str:
        """
        Latest assignment to the Set in GAMS

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1','i2'])
        >>> j = gp.Alias(m, "j", alias_with=i)
        >>> j['i1'] = False
        >>> j.getAssignment()
        'j("i1") = no;'

        """
        if self._assignment is None:
            raise ValidationError("Set was not assigned!")

        return self._assignment.getDeclaration()



from __future__ import annotations

import os
from dataclasses import dataclass, field

import gamspy.utils as utils

__all__ = ["SolverInfo", "add_solver_entry", "remove_solver_entry"]


@dataclass
class SolverInfo:
    solver_id: str
    file_type: str
    dict_type: str
    lic_codes: str
    default_ok_flag: str
    hidden_flag: str
    lines_to_follow: str
    model_types: list = field(default_factory=list)

    def __str__(self) -> str:
        return (
            f"{self.solver_id} {self.file_type} {self.dict_type} "
            f"{self.lic_codes} {self.default_ok_flag} {self.hidden_flag} "
            f"{self.lines_to_follow} {' '.join(self.model_types)}"
        )


def add_solver_entry(system_directory: str, solver_name: str, verbatims: list[str]):
    capabilities_path = os.path.join(system_directory, utils.CAPABILITIES_FILE)
    installed_solvers = utils.getInstalledSolvers(system_directory)
    if solver_name.upper() in installed_solvers:
        print(f"`{solver_name}` already exists in the capabilities file, skipping")
        return

    with open(capabilities_path, encoding="utf-8") as f:
        string = f.read()

    for verbatim in verbatims:
        string = f"{verbatim}\n\n{string}"

        with open(capabilities_path, "w", encoding="utf-8") as f:
            f.write(string)


def find_bounds(system_directory: str, solver_name: str) -> tuple[int, int]:
    capabilities_path = os.path.join(system_directory, utils.CAPABILITIES_FILE)
    with open(capabilities_path, encoding="utf-8") as file:
        lines = file.readlines()

    start_idx = 0
    line_count = 0
    while True:
        line = lines.pop(0)
        if line.startswith("*") or line in {"", "\n"}:
            start_idx += 1
            continue
        if line == "DEFAULTS":
            break

        start_idx += 1
        solver, _, _, _, _, _, num_lines, *_ = line.split()
        for _ in range(int(num_lines) + 2):
            _ = lines.pop(0)

        start_idx += int(num_lines) + 2

        if solver == solver_name.upper():
            line_count = int(num_lines)
            start_idx -= int(num_lines) + 3
            break

    return start_idx, line_count


def remove_solver_entry(system_directory: str, solver_name: str):
    capabilities_path = os.path.join(system_directory, utils.CAPABILITIES_FILE)
    installed_solvers = utils.getInstalledSolvers(system_directory)

    if solver_name.upper() not in installed_solvers:
        print("Solver is not in the capabilities file, skipping")
        return

    line_num, line_count = find_bounds(system_directory, solver_name)

    with open(capabilities_path, encoding="utf-8") as f:
        lines = f.readlines()

    for _ in range(line_count + 3):
        lines.pop(line_num)

    with open(capabilities_path, "w", encoding="utf-8") as f:
        f.writelines(lines)



from __future__ import annotations

import os
import subprocess

import typer

import gamspy.utils as utils
from gamspy.exceptions import ValidationError

app = typer.Typer(
    rich_markup_mode="rich",
    short_help="To probe a node's information.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy probe -j <output_path.json>",
    context_settings={"help_option_names": ["-h", "--help"]},
)


def _probe(json_out: str | None) -> None:
    gamspy_base_dir = utils._get_gamspy_base_directory()
    process = subprocess.run(
        [os.path.join(gamspy_base_dir, "gamsprobe")],
        text=True,
        capture_output=True,
    )

    if process.returncode:
        raise ValidationError(process.stderr)

    print(process.stdout)

    if json_out:
        with open(json_out, "w") as file:
            file.write(process.stdout)

    raise typer.Exit()


@app.callback()
def callback(
    json_out: str | None = typer.Option(
        None,
        "--json-out",
        "-j",
        help="Output path for the JSON file.",
        callback=_probe,
    ),
) -> None: ...


if __name__ == "__main__":
    app()



from __future__ import annotations

import math
from typing import TYPE_CHECKING, Literal

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view

import gamspy as gp
import gamspy.formulations.utils as utils
from gamspy.exceptions import ValidationError
from gamspy.math import dim

if TYPE_CHECKING:
    from gamspy import Parameter, Variable


class Conv2d:
    """
    Formulation generator for 2D Convolution symbol in GAMS. It can
    be used to embed convolutional layers of trained neural networks
    in your problem. It can also be used to embed convolutional layers
    when you need weights as variables.

    Parameters
    ----------
    container : Container
        Container that will contain the new variable and equations.
    in_channel : int
        Number of channels in the input
    out_channel : int
        Number of channels in the output
    kernel_size : int | tuple[int, int]
        Filter size
    stride : int | tuple[int, int]
        Stride in the convolution, by default 1
    padding : int | tuple[int, int] | Literal["same", "valid"]
        Specifies the amount of padding to apply to the input, by default 0.
        If an integer is provided, that padding is applied to both the height and width.
        If a tuple of two integers is given, the first value determines the padding for the
        top and bottom, while the second value sets the padding for the left and right.
        It is also possible to provide string literals "same" and "valid". "same" pads
        the input so the output has the shape as the input. "valid" is the same as no
        padding.
    bias : bool
        Is bias added after the convolution, by default True
    name_prefix : str | None
        Prefix for names of the GAMS symbols generated, by default None which means
        random prefix. Using same name_prefix in different formulations causes name
        conflicts. Do not use same name_prefix again.

    Examples
    --------
    >>> import gamspy as gp
    >>> import numpy as np
    >>> from gamspy.math import dim
    >>> w1 = np.random.rand(2, 1, 3, 3)
    >>> b1 = np.random.rand(2)
    >>> m = gp.Container()
    >>> # in_channels=1, out_channels=2, kernel_size=3x3
    >>> conv1 = gp.formulations.Conv2d(m, 1, 2, 3)
    >>> conv1.load_weights(w1, b1)
    >>> # 10 images, 1 channel, 24 by 24
    >>> inp = gp.Variable(m, domain=dim((10, 1, 24, 24)))
    >>> out, eqs = conv1(inp)
    >>> type(out)
    <class 'gamspy._symbols.variable.Variable'>
    >>> [len(x) for x in out.domain]
    [10, 2, 22, 22]

    """

    def __init__(
        self,
        container: gp.Container,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, int],
        stride: int | tuple[int, int] = 1,
        padding: int | tuple[int, int] | Literal["same", "valid"] = 0,
        bias: bool = True,
        name_prefix: str | None = None,
    ):
        if not (isinstance(in_channels, int) and in_channels > 0):
            raise ValidationError("in_channels must be a positive integer")

        if not (isinstance(out_channels, int) and out_channels > 0):
            raise ValidationError("out_channels must be a positive integer")

        _kernel_size = utils._check_tuple_int(kernel_size, "kernel_size")
        _stride = utils._check_tuple_int(stride, "stride")

        if isinstance(padding, str):
            if padding not in {"same", "valid"}:
                raise ValidationError(
                    "padding must be 'same' or 'valid' when it is a string"
                )

            if padding == "same" and _stride != (1, 1):
                raise ValidationError("'same' padding can only be used with stride=1")

            _padding: tuple[int, int, int, int] | str = (
                (0, 0, 0, 0) if padding == "valid" else "same"
            )

        else:
            _padding = utils._check_padding(padding)

        if not isinstance(bias, bool):
            raise ValidationError("bias must be a boolean")

        self.container = container
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _kernel_size
        self.stride = _stride
        self.padding = _padding
        self.use_bias = bias

        self._state = 0
        self.weight: Parameter | Variable | None = None
        self.weight_array = None
        self.bias: Parameter | Variable | None = None
        self.bias_array = None

        if name_prefix is None:
            name_prefix = gp.utils._get_unique_name()

        self._name_prefix = name_prefix

    def _propagate_bounds(self, input, output, weight, bias, stride, padding):
        # Extract input bounds
        input_bounds = gp.Parameter(
            self.container,
            domain=dim([2, *input.shape]),
            name=utils._generate_name("p", self._name_prefix, "input_bounds"),
        )
        # lower bounds
        input_bounds[("0",) + tuple(input.domain)] = input.lo[...]
        # upper bounds
        input_bounds[("1",) + tuple(input.domain)] = input.up[...]

        # If the bounds are all zeros (None in GAMSPy parameters);
        # we skip matrix multiplication as it will result in zero values
        if input_bounds.records is None:
            out_bounds_array = np.zeros(output.shape)

            if self.use_bias:
                b = self.bias_array[:, np.newaxis, np.newaxis]
                out_bounds_array = out_bounds_array + b

            out_bounds = gp.Parameter(
                self.container,
                domain=dim(output.shape),
                records=out_bounds_array,
                name=utils._generate_name("p", self._name_prefix, "output_bounds"),
            )
            output.lo[...] = out_bounds
            output.up[...] = out_bounds

            return

        input_lower, input_upper = input_bounds.toDense()

        # Check if the input bounds contain (-)infinity values
        inf_exists = input_bounds.countNegInf() or input_bounds.countPosInf()
        out_arr_dtype = np.complex128 if inf_exists else np.float64

        if inf_exists:
            # Encode infinity values as complex numbers
            input_lower = utils._encode_infinity(input_lower)
            input_upper = utils._encode_infinity(input_upper)

        batch, out_channels, h_out, w_out = output.shape
        _, in_channels, h_k, w_k = weight.shape
        stride_y, stride_x = stride
        pad_top, pad_left, pad_bottom, pad_right = padding

        # if any side of the padding is non-zero, we need to pad the input
        if any(padding):
            # Pad the input bounds
            input_lower = np.pad(
                input_lower,
                ((0, 0), (0, 0), (pad_top, pad_bottom), (pad_left, pad_right)),
                mode="constant",
                constant_values=0,
            )
            input_upper = np.pad(
                input_upper,
                ((0, 0), (0, 0), (pad_top, pad_bottom), (pad_left, pad_right)),
                mode="constant",
                constant_values=0,
            )

        # ----- Sliding window view -----
        # Create sliding windows for the input, where each window has the same shape as the kernel.
        # This is done separately for the lower and upper bounds of the input.
        # Then, downsample the sliding windows by selecting every `stride_y` step along the height (vertical) axis
        # and every `stride_x` step along the width (horizontal) axis. This reduces the number of windows
        # by skipping positions based on the user-specified strides.
        #
        # After slicing, the axes are transposed to reorder the dimensions. Specifically:
        # - The original window positions (height and width axes) are moved to the end.
        # - The window content (channel and spatial dimensions) is brought forward.
        #
        # This ensures that the windows are processed in a left-to-right, top-to-bottom order,
        # prioritizing horizontal traversal first, which is the desired behavior.

        windows_lower = sliding_window_view(input_lower, (batch, in_channels, h_k, w_k))
        windows_upper = sliding_window_view(input_upper, (batch, in_channels, h_k, w_k))

        windows_lower = windows_lower[:, :, ::stride_y, ::stride_x, :, :, :].transpose(
            0, 1, 4, 2, 3, 5, 6, 7
        )
        windows_upper = windows_upper[:, :, ::stride_y, ::stride_x, :, :, :].transpose(
            0, 1, 4, 2, 3, 5, 6, 7
        )

        # # Reshape windows for vectorized computation
        windows_lower = windows_lower.reshape(
            batch, h_out, w_out, in_channels, h_k, w_k
        )
        windows_upper = windows_upper.reshape(
            batch, h_out, w_out, in_channels, h_k, w_k
        )

        # Split weights into positive and negative parts
        pos_weight = np.maximum(weight, 0)
        neg_weight = np.minimum(weight, 0)

        # Initialize output bounds
        output_lower = np.zeros(
            (batch, out_channels, h_out, w_out), dtype=out_arr_dtype
        )
        output_upper = np.zeros(
            (batch, out_channels, h_out, w_out), dtype=out_arr_dtype
        )

        if bias is None:
            bias = np.zeros(out_channels)

        for c_out in range(out_channels):
            # Lower bound: sum(input_lower * pos_weight + input_upper * neg_weight)
            output_lower[:, c_out, :, :] = (
                np.sum(
                    windows_lower * pos_weight[c_out, np.newaxis, np.newaxis, :, :, :],
                    axis=(3, 4, 5),
                )
                + np.sum(
                    windows_upper * neg_weight[c_out, np.newaxis, np.newaxis, :, :, :],
                    axis=(3, 4, 5),
                )
                + bias[c_out]
            )

            # Upper bound: sum(input_lower * neg_weight + input_upper * pos_weight)
            output_upper[:, c_out, :, :] = (
                np.sum(
                    windows_lower * neg_weight[c_out, np.newaxis, np.newaxis, :, :, :],
                    axis=(3, 4, 5),
                )
                + np.sum(
                    windows_upper * pos_weight[c_out, np.newaxis, np.newaxis, :, :, :],
                    axis=(3, 4, 5),
                )
                + bias[c_out]
            )

        if inf_exists:
            # Decode complex numbers back to real numbers if infinity values were used
            output_lower = utils._decode_complex_array(output_lower)
            output_upper = utils._decode_complex_array(output_upper)

        out_bounds_array = np.stack([output_lower, output_upper], axis=0)

        out_bounds = gp.Parameter(
            self.container,
            domain=dim([2, *output.shape]),
            records=out_bounds_array,
            name=utils._generate_name("p", self._name_prefix, "output_bounds"),
        )

        output.lo[...] = out_bounds[("0",) + tuple(output.domain)]
        output.up[...] = out_bounds[("1",) + tuple(output.domain)]

    def make_variable(self, *, init_weights=False) -> None:
        """
        Mark Conv2d as variable. After this is called `load_weights`
        cannot be called. Use this when you need to learn the weights
        of your convolutional layer in your optimization model.

        Parameters
        ----------
        init_weights : Optional[bool]
               False by default.
               Whether to initialize weights. It is suggested you set
               this to True unless you want to initialize weights yourself.
               When `init_weights` is set to True, values are initialized from
               :math:`\\mathcal{U}(-\\sqrt{k},\\sqrt{k})`, where :math:`k = 1/[C_{in} * \\prod_{i=0}^{1}{kernel\\_size_n}]`.
        """
        if self._state == 1:
            raise ValidationError(
                "make_variable cannot be used after calling load_weights"
            )

        expected_shape = (
            self.out_channels,
            self.in_channels,
            self.kernel_size[0],
            self.kernel_size[1],
        )

        sk = math.sqrt(
            1 / (self.in_channels * self.kernel_size[0] * self.kernel_size[1])
        )
        if self.weight is None:
            self.weight = gp.Variable(
                self.container,
                domain=dim(expected_shape),
                name=utils._generate_name("v", self._name_prefix, "weight"),
            )
            if init_weights:
                self.weight.l[...] = gp.math.uniform(-sk, sk)

        if self.use_bias and self.bias is None:
            self.bias = gp.Variable(
                self.container,
                domain=dim([self.out_channels]),
                name=utils._generate_name("v", self._name_prefix, "bias"),
            )
            if init_weights:
                self.bias.l[...] = gp.math.uniform(-sk, sk)

        self._state = 2

    def load_weights(self, weight: np.ndarray, bias: np.ndarray | None = None) -> None:
        """
        Mark Conv2d as parameter and load weights from NumPy arrays.
        After this is called `make_variable` cannot be called. Use this
        when you already have the weights of your convolutional layer.

        Parameters
        ----------
        weight : np.ndarray
                 Conv2d layer weights in shape
                 (out_channels x in_channels x kernel_size[0] x kernel_size[1])
        bias : np.ndarray | None
               Conv2d layer bias in shape (out_channels, ), only required when
               bias=True during initialization

        """
        if self._state == 2:
            raise ValidationError(
                "load_weights cannot be used after calling make_variable"
            )

        if self.use_bias is False and bias is not None:
            raise ValidationError(
                "bias must be None since bias was set to False during initialization"
            )

        if self.use_bias is True and bias is None:
            raise ValidationError("bias must be provided")

        if len(weight.shape) != 4:
            raise ValidationError(
                f"expected 4D input for weight (got {len(weight.shape)}D input)"
            )

        expected_shape = (
            self.out_channels,
            self.in_channels,
            self.kernel_size[0],
            self.kernel_size[1],
        )
        if weight.shape != expected_shape:
            raise ValidationError(f"weight expected to be {expected_shape}")

        if bias is not None:
            if len(bias.shape) != 1:
                raise ValidationError(
                    f"expected 1D input for bias (got {len(bias.shape)}D input)"
                )

            if bias.shape != (self.out_channels,):
                raise ValidationError(f"bias expected to be ({self.out_channels},)")

        if self.weight is None:
            self.weight = gp.Parameter(
                self.container,
                domain=dim(expected_shape),
                records=weight,
                name=utils._generate_name("p", self._name_prefix, "weight"),
            )
        else:
            self.weight.setRecords(weight)

        self.weight_array = weight

        if self.use_bias:
            if self.bias is None:
                self.bias = gp.Parameter(
                    self.container,
                    domain=dim([self.out_channels]),
                    records=bias,
                    name=utils._generate_name("p", self._name_prefix, "bias"),
                )
            else:
                self.bias.setRecords(bias)

            self.bias_array = bias

        self._state = 1

    def __call__(
        self, input: gp.Parameter | gp.Variable, propagate_bounds: bool = True
    ) -> tuple[gp.Variable, list[gp.Equation]]:
        """
        Forward pass your input, generate output and equations required for
        calculating the convolution. If `propagate_bounds` is True,
        the `input` is of type variable, and `load_weights` was called, then
        the bounds of the input are propagated to the output.

        Parameters
        ----------
        input : gp.Parameter | gp.Variable
                input to the conv layer, must be in shape
                (batch x in_channels x height x width)
        propagate_bounds : bool = True
                If True, propagate bounds of the input to the output.
                Otherwise, the output variable is unbounded.

        """
        if not isinstance(propagate_bounds, bool):
            raise ValidationError("Expected a boolean for propagate_bounds")

        if self.weight is None:
            raise ValidationError(
                "You must call load_weights or make_variable first before using the Conv2d"
            )

        if len(input.domain) != 4:
            raise ValidationError(f"expected 4D input (got {len(input.domain)}D input)")

        N, C_in, H_in, W_in = input.domain

        if len(C_in) != self.in_channels:
            raise ValidationError("in_channels does not match")

        h_in = len(H_in)
        w_in = len(W_in)

        h_out, w_out = utils._calc_hw(
            self.padding, self.kernel_size, self.stride, h_in, w_in
        )

        out = gp.Variable(
            self.container,
            domain=dim([len(N), self.out_channels, h_out, w_out]),
            name=utils._generate_name("v", self._name_prefix, "output"),
        )

        N, C_out, H_out, W_out = out.domain

        set_out = gp.Equation(
            self.container,
            domain=out.domain,
            name=utils._generate_name("e", self._name_prefix, "set_output"),
        )

        if isinstance(self.padding, str):
            padding = utils._calc_same_padding_2d(self.kernel_size)
        else:
            padding = self.padding

        # expr must have domain N, C_out, H_out, W_out
        top_index = (self.stride[0] * (gp.Ord(H_out) - 1)) - padding[0] + 1
        left_index = (self.stride[1] * (gp.Ord(W_out) - 1)) - padding[1] + 1

        _, _, Hf, Wf = self.weight.domain
        C_in, Hf, Wf, H_in, W_in = utils._next_domains(
            [C_in, Hf, Wf, H_in, W_in],
            out.domain,
        )

        subset = gp.Set(
            self.container,
            domain=[H_out, W_out, Hf, Wf, H_in, W_in],
            name=utils._generate_name("s", self._name_prefix, "conv_subset"),
        )
        subset[
            H_out,
            W_out,
            Hf,
            Wf,
            H_in,
            W_in,
        ].where[
            (gp.Ord(H_in) == (top_index + gp.Ord(Hf) - 1))
            & (gp.Ord(W_in) == (left_index + gp.Ord(Wf) - 1))
        ] = True

        expr = gp.Sum(
            [C_in],
            gp.Sum(
                subset[H_out, W_out, Hf, Wf, H_in, W_in],
                input[N, C_in, H_in, W_in] * self.weight[C_out, C_in, Hf, Wf],
            ),
        )

        if self.use_bias:
            assert self.bias is not None
            expr = expr + self.bias[C_out]

        set_out[N, C_out, H_out, W_out] = out[N, C_out, H_out, W_out] == expr

        # If propagate_bounds is True, weight is a parameter and input is a variable,
        # we will propagate the bounds of the input to the output
        if propagate_bounds and self._state == 1 and isinstance(input, gp.Variable):
            self._propagate_bounds(
                input,
                out,
                self.weight_array,
                self.bias_array,
                self.stride,
                padding,
            )

        return out, [set_out]



from __future__ import annotations

import os
import platform
import subprocess
import sys
from enum import Enum
from pathlib import Path  # noqa: TC003
from typing import Annotated, Optional

import typer

from gamspy.exceptions import ValidationError

app = typer.Typer(
    rich_markup_mode="rich",
    short_help="To run your model with GAMS MIRO.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy run miro [--path <path_to_miro>] [--model <path_to_model>]",
    context_settings={"help_option_names": ["-h", "--help"]},
)


class ModeEnum(str, Enum):
    config = "config"
    base = "base"
    deploy = "deploy"


@app.command(
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy run miro [--path <path_to_miro>] [--model <path_to_model>]",
    short_help="Runs a GAMSPY model with GAMS MIRO app.",
)
def miro(
    mode: Annotated[
        ModeEnum,
        typer.Option("--mode", "-m", help="Execution mode of MIRO"),
    ] = ModeEnum.base,  # type: ignore
    path: Annotated[
        Optional[Path],  # noqa: UP045
        typer.Option(
            "--path",
            "-p",
            exists=True,
            help="Path to the MIRO executable (.exe, .app, or .AppImage)",
        ),
    ] = None,
    skip_execution: Annotated[
        bool,
        typer.Option("--skip-execution", help="Whether to skip model execution."),
    ] = False,
    model: Annotated[
        Optional[Path],  # noqa: UP045,
        typer.Option("--model", "-g", exists=True, help="Path to the GAMSPy model."),
    ] = None,
    args: Annotated[
        list[str] | None,
        typer.Argument(help="Arguments to the GAMSPy model."),
    ] = None,
) -> None:
    if model is None:
        typer.echo("--model must be provided to run MIRO", file=sys.stderr)
        sys.exit(1)

    model = os.path.abspath(model)
    execution_mode = mode.value
    path = os.getenv("MIRO_PATH", path)

    if path is None:
        path = path if path is not None else discover_miro()

    if path is None:
        raise ValidationError("--path must be provided to run MIRO")

    if platform.system() == "Darwin" and os.path.splitext(path)[1] == ".app":
        path = os.path.join(path, "Contents", "MacOS", "GAMS MIRO")

    # Initialize MIRO
    if not skip_execution:
        subprocess_env = os.environ.copy()
        subprocess_env["MIRO"] = "1"
        command = [sys.executable, model]
        if args is not None:
            command += args

        process = subprocess.run(
            command,
            env=subprocess_env,
            capture_output=True,
            text=True,
            encoding="utf-8",
        )
        if process.returncode != 0:
            typer.echo(process.stderr, file=sys.stderr)
            sys.exit(process.returncode)

    # Run MIRO
    subprocess_env = os.environ.copy()
    if execution_mode == "deploy":
        subprocess_env["MIRO_BUILD"] = "true"
        execution_mode = "base"

    subprocess_env["MIRO_MODEL_PATH"] = model
    subprocess_env["MIRO_MODE"] = execution_mode
    subprocess_env["MIRO_DEV_MODE"] = "true"
    subprocess_env["MIRO_USE_TMP"] = "false"
    subprocess_env["PYTHON_EXEC_PATH"] = sys.executable

    subprocess.run([path], env=subprocess_env, check=True)


def discover_miro():
    system = platform.system()
    if system == "Linux":
        return None

    home = os.path.expanduser("~")
    standard_locations = {
        "Darwin": [
            os.path.join(
                "/",
                "Applications",
                "GAMS MIRO.app",
                "Contents",
                "MacOS",
                "GAMS MIRO",
            ),
            os.path.join(
                home,
                "Applications",
                "GAMS MIRO.app",
                "Contents",
                "MacOS",
                "GAMS MIRO",
            ),
        ],
        "Windows": [
            os.path.join("C:\\", "Program Files", "GAMS MIRO", "GAMS MIRO.exe"),
            os.path.join(
                home,
                "AppData",
                "Local",
                "Programs",
                "GAMS MIRO",
                "GAMS MIRO.exe",
            ),
        ],
    }

    if system in ("Darwin", "Windows"):
        for location in standard_locations[system]:
            if os.path.isfile(location):
                return location

    return None


if __name__ == "__main__":
    app()



from __future__ import annotations

import os
import platform
import subprocess
from pathlib import Path  # noqa: TC003
from typing import Annotated

import gamspy_base
import typer

# Valid values for categorical parameters based on help string
VALID_YN = {"0", "N", "1", "Y"}
VALID_DUPLICATES = {"NOCHECK", "ADD", "IGNORE", "ERROR"}
VALID_ORIGNAMES = {"NO", "MODIFIED", "ALL"}
VALID_CONVERTSENSE = {"0", "N", "1", "Y", "MIN", "-1", "MAX"}


# Completion functions for Typer
def complete_yn(ctx: typer.Context, incomplete: str):
    return [v for v in VALID_YN if v.lower().startswith(incomplete.lower())]


def complete_duplicates(ctx: typer.Context, incomplete: str):
    return [v for v in VALID_DUPLICATES if v.lower().startswith(incomplete.lower())]


def complete_orignames(ctx: typer.Context, incomplete: str):
    return [v for v in VALID_ORIGNAMES if v.lower().startswith(incomplete.lower())]


def complete_convertsense(ctx: typer.Context, incomplete: str):
    return [v for v in VALID_CONVERTSENSE if v.lower().startswith(incomplete.lower())]


def mps2gms(
    input_file: Annotated[
        Path, typer.Argument(help="MPS or LP file to translate.", exists=True)
    ],
    gdx_file: Annotated[
        Path | None, typer.Argument(help="Name of GDX output file.")
    ] = None,
    gms_file: Annotated[
        Path | None, typer.Argument(help="Name of GAMS program output file.")
    ] = None,
    py_file: Annotated[
        str | None, typer.Option("--py", help="Name of GAMSPy program output file.")
    ] = None,
    dec_file: Annotated[
        str | None, typer.Option("--dec", help="DEC file for decomposition info.")
    ] = None,
    column_int_vars_binary: Annotated[
        str | None,
        typer.Option(
            "--columnintvarsarebinary",
            help="Integer variables appearing first are binary.",
            autocompletion=complete_yn,
        ),
    ] = None,
    duplicates: Annotated[
        str | None,
        typer.Option(
            help="How to handle multiple coefficients.",
            autocompletion=complete_duplicates,
        ),
    ] = None,
    orignames: Annotated[
        str | None,
        typer.Option(
            help="Whether to make original names available.",
            autocompletion=complete_orignames,
        ),
    ] = None,
    stageshift: Annotated[
        int | None, typer.Option(help="Shift block numbers by this integer.")
    ] = None,
    convertsense: Annotated[
        str | None,
        typer.Option(
            help="Convert the objective function sense.",
            autocompletion=complete_convertsense,
        ),
    ] = None,
):
    """
    Translates an MPS or LP file into equivalent generic GAMS and GAMSPy programs.
    Defaults to writing .py and .gdx files if no outputs are specified.
    """
    binary_name = "mps2gms.exe" if platform.system() == "Windows" else "mps2gms"
    MPS2GMS_PATH = os.path.join(gamspy_base.directory, binary_name)

    if not os.path.exists(MPS2GMS_PATH):
        typer.echo(f"Binary not found: {MPS2GMS_PATH}", err=True)
        raise typer.Exit(code=1)

    # Determine default names based on input stem
    input_name = input_file.with_suffix("")
    actual_gdx = gdx_file if gdx_file else f"{input_name}.gdx"
    actual_py = py_file if py_file else f"{input_name}.py"

    # mps2gms <input> <gdx>
    cmd = [MPS2GMS_PATH, str(input_file), str(actual_gdx)]

    # If gms_file is provided, add it as the third positional;
    # otherwise, explicitly disable GMS output to prioritize py/gdx.
    if gms_file:
        cmd.append(str(gms_file))
    else:
        cmd.append("GMS=")

    # Construct key=value parameters for the binary call
    params = {
        "PY": actual_py,
        "DEC": dec_file,
        "COLUMNINTVARSAREBINARY": column_int_vars_binary,
        "DUPLICATES": duplicates,
        "ORIGNAMES": orignames,
        "STAGESHIFT": stageshift,
        "CONVERTSENSE": convertsense,
    }

    for key, value in params.items():
        if value is not None:
            cmd.append(f"{key}={value}")

    print(f"Running command: {' '.join(cmd)}")
    result = subprocess.run(cmd, text=True)
    raise typer.Exit(code=result.returncode)



from __future__ import annotations

import os
import weakref
from typing import TYPE_CHECKING

from gams.core.gdx import delete_intp, doubleArray, intp_value
from gams.core.gmd import (
    GMD_NRRECORDS,
    GMS_SSSIZE,
    GMS_SV_UNDEF,
    dt_equ,
    dt_par,
    dt_set,
    dt_var,
    gmdAddSymbolPy,
    gmdCreateD,
    gmdFree,
    gmdGetLastError,
    gmdSetSpecialValues,
    gmdSymbolInfo,
    gmdWriteGDX,
    new_gmdHandle_tp,
    new_intp,
)

import gamspy.utils as utils
from gamspy.exceptions import GamspyException

if TYPE_CHECKING:
    from gamspy._workspace import Workspace

SV_UNDEF = GMS_SV_UNDEF
SV_EPS = 4.94066e-324

_spec_values = doubleArray(5)
_spec_values[0] = SV_UNDEF
_spec_values[1] = float("nan")
_spec_values[2] = float("inf")
_spec_values[3] = float("-inf")
_spec_values[4] = SV_EPS


def _int_value_and_free(intP):
    intp_val = intp_value(intP)
    delete_intp(intP)
    return intp_val


class GamsSymbol:
    """
    Representation of a symbol in GAMS. It exists in a Database and contains
    GamsSymbolRecords which one can iterate through.
    """

    def __init__(
        self,
        database: Database,
        name: str,
        dimension: int,
        explanatory_text: str,
    ):
        self.database = database
        self.name = name
        self.dimension = dimension
        self.text = explanatory_text
        self.database.symbols[self.name] = self

    def __len__(self):
        return self.get_number_records()

    def get_number_records(self):
        ret = gmdSymbolInfo(self.database.gmd, self.sym_ptr, GMD_NRRECORDS)
        self.database._check_for_gmd_error(ret[0], self.database.workspace)
        return ret[1]


class GamsSet(GamsSymbol):
    """Representation of a set symbol in GAMS"""

    def __init__(
        self,
        database: Database,
        name: str,
        dimension: int,
        explanatory_text: str = "",
        settype: int = 0,
    ):
        super().__init__(database, name, dimension, explanatory_text)

        self.settype = settype
        rc = new_intp()
        self.sym_ptr = gmdAddSymbolPy(
            self.database.gmd,
            self.name,
            self.dimension,
            dt_set,
            self.settype,
            self.text,
            rc,
        )
        self.database._check_for_gmd_error(
            _int_value_and_free(rc), self.database.workspace
        )


class GamsParameter(GamsSymbol):
    """Representation of a parameter symbol in GAMS"""

    def __init__(
        self,
        database: Database,
        name: str,
        dimension: int,
        explanatory_text: str = "",
    ):
        super().__init__(database, name, dimension, explanatory_text)

        rc = new_intp()
        self.sym_ptr = gmdAddSymbolPy(
            self.database.gmd,
            self.name,
            self.dimension,
            dt_par,
            0,
            self.text,
            rc,
        )
        self.database._check_for_gmd_error(
            _int_value_and_free(rc), self.database.workspace
        )


class GamsVariable(GamsSymbol):
    """Representation of a variable symbol in GAMS"""

    def __init__(
        self,
        database: Database,
        name: str,
        dimension: int,
        vartype: int,
        explanatory_text: str = "",
    ):
        super().__init__(database, name, dimension, explanatory_text)

        self.vartype = vartype
        rc = new_intp()
        self.sym_ptr = gmdAddSymbolPy(
            self.database.gmd,
            self.name,
            self.dimension,
            dt_var,
            self.vartype,
            self.text,
            rc,
        )
        self.database._check_for_gmd_error(
            _int_value_and_free(rc), self.database.workspace
        )


class GamsEquation(GamsSymbol):
    """Representation of an equation symbol in GAMS"""

    def __init__(
        self,
        database: Database,
        name: str,
        dimension: int,
        equtype: int,
        explanatory_text: str = "",
    ):
        super().__init__(database, name, dimension, explanatory_text)

        self.equtype = equtype
        rc = new_intp()
        self.sym_ptr = gmdAddSymbolPy(
            self.database.gmd,
            self.name,
            self.dimension,
            dt_equ,
            self.equtype,
            self.text,
            rc,
        )
        self.database._check_for_gmd_error(
            _int_value_and_free(rc), self.database.workspace
        )


class Database:
    """Communicates data between Python and GAMS"""

    def __init__(self, ws: Workspace):
        self.symbols: dict = {}
        self.workspace = ws
        self.gmd = new_gmdHandle_tp()

        ret = gmdCreateD(self.gmd, utils._get_gamspy_base_directory(), GMS_SSSIZE)
        if not ret[0]:
            raise GamspyException(ret[1])

        rc = gmdSetSpecialValues(self.gmd, _spec_values)
        self._check_for_gmd_error(rc, self.workspace)

        weakref.finalize(self, self.cleanup, self.gmd)

    @staticmethod
    def cleanup(gmd):
        gmdFree(gmd)

    def __len__(self):
        return len(self.symbols)

    def __getitem__(self, symbol_name: str):
        return self.symbols[symbol_name]

    def _check_for_gmd_error(self, rc, workspace: Workspace):
        if not rc:
            msg = gmdGetLastError(self.gmd)[1]
            workspace._errors.append(msg)
            raise GamspyException(msg, rc)

    def add_equation(
        self,
        name: str,
        dimension: int,
        equtype: int,
        explanatory_text: str = "",
    ) -> GamsEquation:
        return GamsEquation(self, name, dimension, equtype, explanatory_text)

    def add_variable(
        self,
        name: str,
        dimension: int,
        vartype: int,
        explanatory_text: str = "",
    ) -> GamsVariable:
        return GamsVariable(self, name, dimension, vartype, explanatory_text)

    def add_set(
        self,
        name: str,
        dimension: int,
        explanatory_text: str = "",
        settype: int = 0,
    ) -> GamsSet:
        return GamsSet(self, name, dimension, explanatory_text, settype=settype)

    def add_parameter(
        self, name: str, dimension: int, explanatory_text: str = ""
    ) -> GamsParameter:
        return GamsParameter(self, name, dimension, explanatory_text)

    def export(self, file_path: str) -> None:
        """Writes database into a GDX file"""
        assert file_path.endswith(".gdx"), (
            f"File path should point to a gdx file but got `{file_path}`"
        )

        if os.path.isabs(file_path):
            rc = gmdWriteGDX(self.gmd, file_path, False)
        else:
            rc = gmdWriteGDX(
                self.gmd,
                os.path.join(self.workspace.working_directory, file_path),
                False,
            )
        self._check_for_gmd_error(rc, self.workspace)
